# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import contextlib

import mock
from neutron.api import extensions
from neutron_dynamic_routing.db import bgp_db  # noqa
from neutron_dynamic_routing import extensions as dr_extensions
from neutron_dynamic_routing.extensions import bgp as ext_bgp
from neutron_dynamic_routing.tests.unit.db import test_bgp_db
from neutron_lib.api.definitions import address_scope
from neutron_lib import context
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory

from vmware_nsx.common import exceptions as exc
from vmware_nsx.db import nsxv_db
from vmware_nsx.plugins.nsx_v.drivers import (
    shared_router_driver as router_driver)
from vmware_nsx.services.dynamic_routing import bgp_plugin
from vmware_nsx.services.dynamic_routing.nsx_v import driver as bgp_driver
from vmware_nsx.tests.unit.nsx_v import test_plugin

BGP_PLUGIN = 'vmware_nsx.services.dynamic_routing.bgp_plugin.NSXvBgpPlugin'


class TestNSXvBgpPlugin(test_plugin.NsxVPluginV2TestCase,
                        test_bgp_db.BgpTests):
    def setUp(self):
        extensions.append_api_extensions_path(dr_extensions.__path__)
        service_plugins = {ext_bgp.BGP_EXT_ALIAS: BGP_PLUGIN}
        super(TestNSXvBgpPlugin, self).setUp(service_plugins=service_plugins)
        self.bgp_plugin = bgp_plugin.NSXvBgpPlugin()
        self.nsxv_driver = self.bgp_plugin.drivers['nsx-v']
        self.nsxv_driver._validate_gateway_network = mock.Mock()
        self.nsxv_driver._validate_bgp_configuration_on_peer_esg = (
            mock.Mock())
        self.plugin = directory.get_plugin()
        self.l3plugin = self.plugin
        self.plugin.init_is_complete = True
        self.context = context.get_admin_context()
        self.project_id = 'dummy_project'

    @contextlib.contextmanager
    def gw_network(self, external=True, **kwargs):
        with super(TestNSXvBgpPlugin, self).gw_network(external=external,
                                                       **kwargs) as gw_net:
            if external:
                gw_net['network']['router:external'] = True
                gw_net['network'][address_scope.IPV4_ADDRESS_SCOPE] = True
            yield gw_net

    @contextlib.contextmanager
    def subnet(self, network=None, **kwargs):
        if network and network['network'].get('router:external'):
            kwargs['gateway_ip'] = None
            kwargs['enable_dhcp'] = False

        with super(TestNSXvBgpPlugin, self).subnet(network=network,
                                                   **kwargs) as sub:
            yield sub

    @contextlib.contextmanager
    def router(self, **kwargs):
        if 'external_gateway_info' in kwargs:
            kwargs['external_gateway_info']['enable_snat'] = False
        with super(TestNSXvBgpPlugin, self).router(**kwargs) as r:
            yield r

    @contextlib.contextmanager
    def esg_bgp_peer(self, esg_id):
        data = {'name': '',
                'peer_ip': '192.168.1.10',
                'remote_as': '65000',
                'esg_id': esg_id,
                'auth_type': 'none',
                'password': '',
                'tenant_id': self.project_id}
        bgp_peer = self.bgp_plugin.create_bgp_peer(self.context,
                                                   {'bgp_peer': data})
        yield bgp_peer
        self.bgp_plugin.delete_bgp_peer(self.context, bgp_peer['id'])

    @contextlib.contextmanager
    def bgp_speaker(self, ip_version, local_as, name='my-speaker',
                    advertise_fip_host_routes=True,
                    advertise_tenant_networks=True,
                    networks=None, peers=None):
        data = {'ip_version': ip_version,
                test_bgp_db.ADVERTISE_FIPS_KEY: advertise_fip_host_routes,
                'advertise_tenant_networks': advertise_tenant_networks,
                'local_as': local_as, 'name': name,
                'tenant_id': self.project_id}
        bgp_speaker = self.bgp_plugin.create_bgp_speaker(self.context,
                                                        {'bgp_speaker': data})
        bgp_speaker_id = bgp_speaker['id']

        if networks:
            for network_id in networks:
                self.bgp_plugin.add_gateway_network(
                                                   self.context,
                                                   bgp_speaker_id,
                                                   {'network_id': network_id})
        if peers:
            for peer_id in peers:
                self.bgp_plugin.add_bgp_peer(self.context, bgp_speaker_id,
                                             {'bgp_peer_id': peer_id})

        yield self.bgp_plugin.get_bgp_speaker(self.context, bgp_speaker_id)

    def test_create_v6_bgp_speaker(self):
        fake_bgp_speaker = {
            "bgp_speaker": {
                "ip_version": 6,
                "local_as": "1000",
                "name": "bgp-speaker",
                "tenant_id": self.project_id
            }
        }
        self.assertRaises(n_exc.InvalidInput,
                          self.bgp_plugin.create_bgp_speaker,
                          self.context, fake_bgp_speaker)

    def test_create_v6_bgp_peer(self):
        fake_bgp_peer = {
            "bgp_peer": {
                "auth_type": "none",
                "remote_as": "1000",
                "name": "bgp-peer",
                "peer_ip": "fc00::/7",
                "tenant_id": self.project_id
            }
        }
        self.assertRaises(n_exc.InvalidInput,
                          self.bgp_plugin.create_bgp_peer,
                          self.context, fake_bgp_peer)

    def test_bgp_peer_esg_id(self):
        edge_id = 'edge-123'
        with self.esg_bgp_peer(esg_id='edge-123') as esg_peer:
            self.assertEqual(edge_id, esg_peer['esg_id'])

            peer_id = esg_peer['id']
            bgp_peer = self.bgp_plugin.get_bgp_peer(self.context, peer_id)
            self.assertEqual(edge_id, bgp_peer['esg_id'])

    def test_create_bgp_peer_md5_auth_no_password(self):
        bgp_peer = {'bgp_peer':
                    {'auth_type': 'md5', 'password': None,
                     'peer_ip': '10.0.0.3',
                     'tenant_id': self.project_id}}
        self.assertRaises(ext_bgp.InvalidBgpPeerMd5Authentication,
                          self.bgp_plugin.create_bgp_peer,
                          self.context, bgp_peer)

    def test_add_non_external_gateway_network(self):
        self.nsxv_driver._validate_gateway_network = (
            bgp_driver.NSXvBgpDriver(
                self.bgp_plugin)._validate_gateway_network)
        with self.gw_network(external=False) as net,\
            self.subnetpool_with_address_scope(4,
                                               prefixes=['8.0.0.0/8']) as sp:
            network_id = net['network']['id']
            with self.bgp_speaker(sp['ip_version'], 1234) as speaker:
                self.assertRaises(exc.NsxBgpNetworkNotExternal,
                                  self.bgp_plugin.add_gateway_network,
                                  self.context, speaker['id'],
                                  {'network_id': network_id})

    @mock.patch.object(nsxv_db, 'get_nsxv_bgp_speaker_binding',
                       return_value={'bgp_identifier': '10.0.0.11'})
    def test_shared_router_on_gateway_clear(self, m1):
        with self.gw_network(external=True) as net,\
            self.subnetpool_with_address_scope(4,
                                               prefixes=['10.0.0.0/24']) as sp:
            with self.subnet(network=net,
                             subnetpool_id=sp['id']) as s1,\
                self.bgp_speaker(sp['ip_version'], 1234,
                                 networks=[net['network']['id']]):
                subnet_id = s1['subnet']['id']
                gw_info1 = {'network_id': net['network']['id'],
                            'external_fixed_ips': [{'ip_address': '10.0.0.11',
                                                    'subnet_id': subnet_id}]}
                gw_info2 = {'network_id': net['network']['id'],
                            'external_fixed_ips': [{'ip_address': '10.0.0.12',
                                                    'subnet_id': subnet_id}]}
                router_obj = router_driver.RouterSharedDriver(self.plugin)
                with mock.patch.object(self.plugin, '_find_router_driver',
                                       return_value=router_obj):
                    with self.router(external_gateway_info=gw_info1) as rtr1,\
                        self.router(external_gateway_info=gw_info2) as rtr2,\
                        mock.patch.object(
                            self.nsxv_driver, '_get_router_edge_info',
                            return_value=('edge-1', False)),\
                        mock.patch.object(
                            self.plugin.edge_manager,
                            'get_routers_on_same_edge',
                            return_value=[rtr1['id'], rtr2['id']]),\
                        mock.patch.object(
                            self.nsxv_driver,
                            '_update_edge_bgp_identifier') as up_bgp:
                        gw_clear = {u'router': {u'external_gateway_info': {}}}
                        self.plugin.update_router(self.context,
                                                  rtr1['id'],
                                                  gw_clear)
                        up_bgp.assert_called_once_with(mock.ANY,
                                                       mock.ANY,
                                                       mock.ANY,
                                                       '10.0.0.12')

    def test__bgp_speakers_for_gateway_network_by_ip_version(self):
        # REVISIT(roeyc): Base class test use ipv6 which is not supported.
        pass

    def test__bgp_speakers_for_gateway_network_by_ip_version_no_binding(self):
        # REVISIT(roeyc): Base class test use ipv6 which is not supported.
        pass

    def test__tenant_prefixes_by_router_no_gateway_port(self):
        # REVISIT(roeyc): Base class test use ipv6 which is not supported.
        pass

    def test_all_routes_by_bgp_speaker_different_tenant_address_scope(self):
        # REVISIT(roeyc): Base class test use ipv6 which is not supported.
        pass

    def test__get_address_scope_ids_for_bgp_speaker(self):
        pass

    def test__get_dvr_fip_host_routes_by_binding(self):
        pass

    def test__get_dvr_fip_host_routes_by_router(self):
        pass

    def test__get_fip_next_hop_dvr(self):
        pass

    def test__get_fip_next_hop_legacy(self):
        pass

    def test_get_routes_by_bgp_speaker_id_with_fip_dvr(self):
        pass

    def test_ha_router_fips_has_no_next_hop_to_fip_agent_gateway(self):
        pass

    def test_legacy_router_fips_has_no_next_hop_to_fip_agent_gateway(self):
        pass

    def test_floatingip_update_callback(self):
        pass

    def test_get_ipv6_tenant_subnet_routes_by_bgp_speaker_ipv6(self):
        pass

    def test_get_routes_by_bgp_speaker_id_with_fip(self):
        # base class tests uses no-snat router with floating ips
        self.skipTest('No SNAT with floating ips not supported')

    def test_get_routes_by_bgp_speaker_binding_with_fip(self):
        # base class tests uses no-snat router with floating ips
        self.skipTest('No SNAT with floating ips not supported')

    def test__get_routes_by_router_with_fip(self):
        # base class tests uses no-snat router with floating ips
        self.skipTest('No SNAT with floating ips not supported')

    def test_add_bgp_peer_with_bad_id(self):
            with self.subnetpool_with_address_scope(
                4, prefixes=['8.0.0.0/8']) as sp:
                with self.bgp_speaker(sp['ip_version'], 1234) as speaker:
                    self.assertRaises(ext_bgp.BgpPeerNotFound,
                                      self.bgp_plugin.add_bgp_peer,
                                      self.context,
                                      speaker['id'],
                                      {'bgp_peer_id': 'aaa'})



# Copyright 2017 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Port vnic_type support

Revision ID: e4c503f4133f
Revises: 01a33f93f5fd
Create Date: 2017-02-20 00:05:30.894680

"""

# revision identifiers, used by Alembic.
revision = 'e4c503f4133f'
down_revision = '01a33f93f5fd'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'nsxv_port_ext_attributes',
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.Column('vnic_type', sa.String(length=64), nullable=False,
                  server_default='normal'),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('port_id'))



# Copyright 2013 VMware, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import xml.etree.ElementTree as et

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import strutils
import six

from vmware_nsx.common import nsxv_constants
from vmware_nsx.common import utils
from vmware_nsx.plugins.nsx_v.vshield.common import constants
from vmware_nsx.plugins.nsx_v.vshield.common import exceptions
from vmware_nsx.plugins.nsx_v.vshield.common import VcnsApiClient

LOG = logging.getLogger(__name__)

HTTP_GET = "GET"
HTTP_POST = "POST"
HTTP_DELETE = "DELETE"
HTTP_PUT = "PUT"
URI_PREFIX = "/api/4.0/edges"

#FwaaS constants
FIREWALL_SERVICE = "firewall/config"
FIREWALL_RULE_RESOURCE = "rules"

#NSXv Constants
FIREWALL_PREFIX = '/api/4.0/firewall/globalroot-0/config'
FIREWALL_REDIRECT_SEC_TYPE = 'layer3redirectsections'
SECURITYGROUP_PREFIX = '/api/2.0/services/securitygroup'
VDN_PREFIX = '/api/2.0/vdn'
SERVICES_PREFIX = '/api/2.0/services'
SPOOFGUARD_PREFIX = '/api/4.0/services/spoofguard'
TRUSTSTORE_PREFIX = '%s/%s' % (SERVICES_PREFIX, 'truststore')
EXCLUDELIST_PREFIX = '/api/2.1/app/excludelist'
SERVICE_INSERTION_PROFILE_PREFIX = '/api/2.0/si/serviceprofile'
SECURITY_POLICY_PREFIX = '/api/2.0/services/policy/securitypolicy'
APPLICATION_PREFIX = '%s/%s' % (SERVICES_PREFIX, 'application')

#LbaaS Constants
LOADBALANCER_SERVICE = "loadbalancer/config"
LOADBALANCER_STATS = "loadbalancer/statistics"
VIP_RESOURCE = "virtualservers"
POOL_RESOURCE = "pools"
MONITOR_RESOURCE = "monitors"
APP_PROFILE_RESOURCE = "applicationprofiles"
APP_RULE_RESOURCE = "applicationrules"

# IPsec VPNaaS Constants
IPSEC_VPN_SERVICE = 'ipsec/config'

# Dhcp constants
DHCP_SERVICE = "dhcp/config"
DHCP_BINDING_RESOURCE = "bindings"

# Syetem control constants
SYSCTL_SERVICE = 'systemcontrol/config'

# L2 gateway constants
BRIDGE = "bridging/config"

# IPAM constants
IPAM_POOL_SCOPE = "scope/globalroot-0"
IPAM_POOL_SERVICE = "ipam/pools"

# Self Signed Certificate constants
CSR = "csr"
CERTIFICATE = "certificate"

NETWORK_TYPES = ['Network', 'VirtualWire', 'DistributedVirtualPortgroup']

# Dynamic routing constants
ROUTING_CONFIG = "routing/config"
BGP_ROUTING_CONFIG = "routing/config/bgp"
ELAPSED_TIME_THRESHOLD = 30
MAX_EDGE_DEPLOY_TIMEOUT = 1200


def retry_upon_exception_exclude_error_codes(
    exc, excluded_errors, delay=0.5, max_delay=4, max_attempts=0):
    if not max_attempts:
        max_attempts = cfg.CONF.nsxv.retries
    return utils.retry_upon_exception_exclude_error_codes(
        exc, excluded_errors, delay, max_delay, max_attempts)


def retry_upon_exception(exc, delay=0.5, max_delay=4, max_attempts=0):
    if not max_attempts:
        max_attempts = cfg.CONF.nsxv.retries
    return utils.retry_upon_exception(exc, delay, max_delay, max_attempts)


class Vcns(object):

    def __init__(self, address, user, password, ca_file, insecure):
        self.address = address
        self.user = user
        self.password = password
        self.ca_file = ca_file
        self.insecure = insecure
        self.jsonapi_client = VcnsApiClient.VcnsApiHelper(
            address, user, password, format='json', ca_file=ca_file,
            insecure=insecure, timeout=cfg.CONF.nsxv.nsx_transaction_timeout)
        self.xmlapi_client = VcnsApiClient.VcnsApiHelper(
            address, user, password, format='xml', ca_file=ca_file,
            insecure=insecure, timeout=cfg.CONF.nsxv.nsx_transaction_timeout)
        self._nsx_version = None
        self._normalized_scoping_objects = None
        self._normalized_global_objects = None

    @retry_upon_exception(exceptions.ServiceConflict)
    def _client_request(self, client, method, uri,
                        params, headers, encodeParams, timeout=None):
        return client(method, uri, params, headers, encodeParams,
                      timeout=timeout)

    def do_request(self, method, uri, params=None, format='json', **kwargs):
        msg = ("VcnsApiHelper('%(method)s', '%(uri)s', '%(body)s')" %
               {'method': method,
                'uri': uri,
                'body': jsonutils.dumps(params)})
        LOG.debug(strutils.mask_password(msg))

        headers = kwargs.get('headers')
        encodeParams = kwargs.get('encode', True)
        if format == 'json':
            _client = self.jsonapi_client.request
        else:
            _client = self.xmlapi_client.request

        timeout = kwargs.get('timeout')
        ts = time.time()
        header, content = self._client_request(_client, method, uri, params,
                                               headers, encodeParams,
                                               timeout=timeout)
        te = time.time()
        elapsed_time = te - ts

        LOG.debug('VcnsApiHelper for %(method)s %(uri)s took %(seconds)2.4f. '
                  'reply: header=%(header)s content=%(content)s',
                  {'method': method, 'uri': uri,
                   'header': header, 'content': content,
                   'seconds': elapsed_time})
        if elapsed_time > ELAPSED_TIME_THRESHOLD:
            LOG.warning('Vcns call for %(method)s %(uri)s took %(seconds)2.4f',
                        {'method': method, 'uri': uri,
                         'seconds': elapsed_time})

        if content == '':
            return header, {}
        if kwargs.get('decode', True):
            content = jsonutils.loads(content)
        return header, content

    def edges_lock_operation(self):
        uri = URI_PREFIX + "?lockUpdatesOnEdge=true"
        return self.do_request(HTTP_POST, uri, decode=False)

    @retry_upon_exception(exceptions.ResourceNotFound)
    @retry_upon_exception(exceptions.RequestBad)
    def deploy_edge(self, request):
        uri = URI_PREFIX
        return self.do_request(HTTP_POST, uri, request, decode=False,
                               timeout=MAX_EDGE_DEPLOY_TIMEOUT)

    def update_edge(self, edge_id, request):
        uri = "%s/%s" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_PUT, uri, request, decode=False)

    def get_edge_id(self, job_id):
        uri = URI_PREFIX + "/jobs/%s" % job_id
        return self.do_request(HTTP_GET, uri, decode=True)

    def get_edge_jobs(self, edge_id):
        uri = URI_PREFIX + "/%s/jobs" % edge_id
        return self.do_request(HTTP_GET, uri, decode=True)

    def get_edge_deploy_status(self, edge_id):
        uri = URI_PREFIX + "/%s/status?getlatest=false" % edge_id
        return self.do_request(HTTP_GET, uri, decode="True")

    def delete_edge(self, edge_id):
        uri = "%s/%s" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_DELETE, uri)

    def add_vdr_internal_interface(self, edge_id, interface):
        uri = "%s/%s/interfaces?action=patch" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_POST, uri, interface, decode=True)

    def get_vdr_internal_interface(self, edge_id, interface_index):
        uri = "%s/%s/interfaces/%s" % (URI_PREFIX, edge_id, interface_index)
        return self.do_request(HTTP_GET, uri, decode=True)

    def update_vdr_internal_interface(self, edge_id,
                                      interface_index, interface):
        uri = "%s/%s/interfaces/%s" % (URI_PREFIX, edge_id, interface_index)
        return self.do_request(HTTP_PUT, uri, interface,
                               format='xml', decode=True)

    @retry_upon_exception(exceptions.RequestBad)
    def delete_vdr_internal_interface(self, edge_id, interface_index):
        uri = "%s/%s/interfaces/%d" % (URI_PREFIX, edge_id, interface_index)
        return self.do_request(HTTP_DELETE, uri, decode=True)

    def get_interfaces(self, edge_id):
        uri = "%s/%s/vnics" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    @retry_upon_exception(exceptions.RequestBad)
    def update_interface(self, edge_id, vnic):
        uri = "%s/%s/vnics/%d" % (URI_PREFIX, edge_id,
                                  vnic['index'])
        return self.do_request(HTTP_PUT, uri, vnic, decode=True)

    def delete_interface(self, edge_id, vnic_index):
        uri = "%s/%s/vnics/%d" % (URI_PREFIX, edge_id, vnic_index)
        return self.do_request(HTTP_DELETE, uri, decode=True)

    def get_nat_config(self, edge_id):
        uri = "%s/%s/nat/config" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    def update_nat_config(self, edge_id, nat):
        uri = "%s/%s/nat/config" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_PUT, uri, nat, decode=True)

    def delete_nat_rule(self, edge_id, rule_id):
        uri = "%s/%s/nat/config/rules/%s" % (URI_PREFIX, edge_id, rule_id)
        return self.do_request(HTTP_DELETE, uri, decode=True)

    def get_edge_status(self, edge_id):
        uri = "%s/%s/status?getlatest=false" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    def get_edge(self, edge_id):
        uri = "%s/%s" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    def _get_edges(self, startindex=0):
        uri = '%s?startIndex=%d' % (URI_PREFIX, startindex)
        return self.do_request(HTTP_GET, uri, decode=True)

    def get_edges(self):
        edges = []
        h, d = self._get_edges()
        edges.extend(d['edgePage']['data'])
        paging_info = d['edgePage']['pagingInfo']
        page_size = int(paging_info['pageSize'])
        count = int(paging_info['totalCount'])
        LOG.debug("There are total %s edges and page size is %s",
                  count, page_size)
        pages = count / page_size + 1
        for i in range(1, pages):
            start_index = page_size * i
            h, d = self._get_edges(start_index)
            edges.extend(d['edgePage']['data'])
        return edges

    def get_edge_syslog(self, edge_id):
        uri = "%s/%s/syslog/config" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    def update_edge_syslog(self, edge_id, config):
        uri = "%s/%s/syslog/config" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_PUT, uri, config)

    def delete_edge_syslog(self, edge_id):
        uri = "%s/%s/syslog/config" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_DELETE, uri)

    def update_edge_config_with_modifier(self, edge_id, module, modifier):
        uri = "%s/%s/%s/config" % (URI_PREFIX, edge_id, module)
        config = self.do_request(HTTP_GET, uri)[1]
        if modifier(config):
            return self.do_request(HTTP_PUT, uri, config)

    def get_edge_interfaces(self, edge_id):
        uri = "%s/%s/interfaces" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    def get_routes(self, edge_id):
        uri = "%s/%s/routing/config/static" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_GET, uri)

    def update_routes(self, edge_id, routes):
        uri = "%s/%s/routing/config/static" % (URI_PREFIX, edge_id)
        return self.do_request(HTTP_PUT, uri, routes)

    def create_lswitch(self, lsconfig):
        uri = "/api/ws.v1/lswitch"
        return self.do_request(HTTP_POST, uri, lsconfig, decode=True)

    def delete_lswitch(self, lswitch_id):
        uri = "/api/ws.v1/lswitch/%s" % lswitch_id
        return self.do_request(HTTP_DELETE, uri)

    def get_loadbalancer_config(self, edge_id):
        uri = self._build_uri_path(edge_id, LOADBALANCER_SERVICE)
        return self.do_request(HTTP_GET, uri, decode=True)

    def get_loadbalancer_statistics(self, edge_id):
        uri = self._build_uri_path(edge_id, LOADBALANCER_STATS)
        return self.do_request(HTTP_GET, uri, decode=True)

    def enable_service_loadbalancer(self, edge_id, config):
        uri = self._build_uri_path(edge_id, LOADBALANCER_SERVICE)
        return self.do_request(HTTP_PUT, uri, config)

    def sync_firewall(self):
        for cluster_id in cfg.CONF.nsxv.cluster_moid:
            uri = '/api/4.0/firewall/forceSync/%s' % cluster_id
            self.do_request(HTTP_POST, uri)

    def update_firewall(self, edge_id, fw_req):
        uri = self._build_uri_path(
            edge_id, FIREWALL_SERVICE)
        return self.do_request(HTTP_PUT, uri, fw_req)

    def delete_firewall(self, edge_id):
        uri = self._build_uri_path(
            edge_id, FIREWALL_SERVICE, None)
        return self.do_request(HTTP_DELETE, uri)

    def update_firewall_rule(self, edge_id, vcns_rule_id, fwr_req):
        uri = self._build_uri_path(
            edge_id, FIREWALL_SERVICE,
            FIREWALL_RULE_RESOURCE,
            vcns_rule_id)
        return self.do_request(HTTP_PUT, uri, fwr_req)

    def delete_firewall_rule(self, edge_id, vcns_rule_id):
        uri = self._build_uri_path(
            edge_id, FIREWALL_SERVICE,
            FIREWALL_RULE_RESOURCE,
            vcns_rule_id)
        return self.do_request(HTTP_DELETE, uri)

    def add_firewall_rule_above(self, edge_id, ref_vcns_rule_id, fwr_req):
        uri = self._build_uri_path(
            edge_id, FIREWALL_SERVICE,
            FIREWALL_RULE_RESOURCE)
        uri += "?aboveRuleId=" + ref_vcns_rule_id
        return self.do_request(HTTP_POST, uri, fwr_req)

    def add_firewall_rule(self, edge_id, fwr_req):
        uri = self._build_uri_path(
            edge_id, FIREWALL_SERVICE,
            FIREWALL_RULE_RESOURCE)
        return self.do_request(HTTP_POST, uri, fwr_req)

    def update_firewall_default_policy(self, edge_id, fw_req):
        uri = self._build_uri_path(
            edge_id, FIREWALL_SERVICE, 'defaultpolicy')
        return self.do_request(HTTP_PUT, uri, fw_req)

    def get_firewall(self, edge_id):
        uri = self._build_uri_path(edge_id, FIREWALL_SERVICE)
        return self.do_request(HTTP_GET, uri, decode=True)

    def get_firewall_rule(self, edge_id, vcns_rule_id):
        uri = self._build_uri_path(
            edge_id, FIREWALL_SERVICE,
            FIREWALL_RULE_RESOURCE,
            vcns_rule_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    #
    #Edge LBAAS call helper
    #
    def create_vip(self, edge_id, vip_new):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            VIP_RESOURCE)
        return self.do_request(HTTP_POST, uri, vip_new)

    def get_vip(self, edge_id, vip_vseid):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            VIP_RESOURCE, vip_vseid)
        return self.do_request(HTTP_GET, uri, decode=True)

    def update_vip(self, edge_id, vip_vseid, vip_new):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            VIP_RESOURCE, vip_vseid)
        return self.do_request(HTTP_PUT, uri, vip_new)

    def delete_vip(self, edge_id, vip_vseid):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            VIP_RESOURCE, vip_vseid)
        return self.do_request(HTTP_DELETE, uri)

    def create_pool(self, edge_id, pool_new):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            POOL_RESOURCE)
        return self.do_request(HTTP_POST, uri, pool_new)

    def get_pool(self, edge_id, pool_vseid):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            POOL_RESOURCE, pool_vseid)
        return self.do_request(HTTP_GET, uri, decode=True)

    def update_pool(self, edge_id, pool_vseid, pool_new):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            POOL_RESOURCE, pool_vseid)
        return self.do_request(HTTP_PUT, uri, pool_new)

    def delete_pool(self, edge_id, pool_vseid):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            POOL_RESOURCE, pool_vseid)
        return self.do_request(HTTP_DELETE, uri)

    def create_health_monitor(self, edge_id, monitor_new):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            MONITOR_RESOURCE)
        return self.do_request(HTTP_POST, uri, monitor_new)

    def get_health_monitor(self, edge_id, monitor_vseid):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            MONITOR_RESOURCE, monitor_vseid)
        return self.do_request(HTTP_GET, uri, decode=True)

    def update_health_monitor(self, edge_id, monitor_vseid, monitor_new):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            MONITOR_RESOURCE,
            monitor_vseid)
        return self.do_request(HTTP_PUT, uri, monitor_new)

    def delete_health_monitor(self, edge_id, monitor_vseid):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            MONITOR_RESOURCE,
            monitor_vseid)
        return self.do_request(HTTP_DELETE, uri)

    def create_app_profile(self, edge_id, app_profile):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            APP_PROFILE_RESOURCE)
        return self.do_request(HTTP_POST, uri, app_profile)

    def update_app_profile(self, edge_id, app_profileid, app_profile):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            APP_PROFILE_RESOURCE, app_profileid)
        return self.do_request(HTTP_PUT, uri, app_profile)

    def delete_app_profile(self, edge_id, app_profileid):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            APP_PROFILE_RESOURCE,
            app_profileid)
        return self.do_request(HTTP_DELETE, uri)

    def create_app_rule(self, edge_id, app_rule):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            APP_RULE_RESOURCE)
        return self.do_request(HTTP_POST, uri, app_rule)

    def update_app_rule(self, edge_id, app_ruleid, app_rule):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            APP_RULE_RESOURCE, app_ruleid)
        return self.do_request(HTTP_PUT, uri, app_rule)

    def delete_app_rule(self, edge_id, app_ruleid):
        uri = self._build_uri_path(
            edge_id, LOADBALANCER_SERVICE,
            APP_RULE_RESOURCE,
            app_ruleid)
        return self.do_request(HTTP_DELETE, uri)

    def update_ipsec_config(self, edge_id, ipsec_config):
        uri = self._build_uri_path(edge_id, IPSEC_VPN_SERVICE)
        return self.do_request(HTTP_PUT, uri, ipsec_config)

    def delete_ipsec_config(self, edge_id):
        uri = self._build_uri_path(edge_id, IPSEC_VPN_SERVICE)
        return self.do_request(HTTP_DELETE, uri)

    def get_ipsec_config(self, edge_id):
        uri = self._build_uri_path(edge_id, IPSEC_VPN_SERVICE)
        return self.do_request(HTTP_GET, uri)

    @retry_upon_exception(exceptions.RequestBad)
    def create_virtual_wire(self, vdn_scope_id, request):
        """Creates a VXLAN virtual wire

        The method will return the virtual wire ID.
        """
        uri = '/api/2.0/vdn/scopes/%s/virtualwires' % vdn_scope_id
        return self.do_request(HTTP_POST, uri, request, format='xml',
                               decode=False)

    def delete_virtual_wire(self, virtualwire_id):
        """Deletes a virtual wire."""
        uri = '/api/2.0/vdn/virtualwires/%s' % virtualwire_id
        return self.do_request(HTTP_DELETE, uri, format='xml')

    def create_port_group(self, dvs_id, request):
        """Creates a port group on a DVS

        The method will return the port group ID.
        """
        uri = '/api/2.0/xvs/switches/%s/networks' % dvs_id
        return self.do_request(HTTP_POST, uri, request, format='xml',
                               decode=False)

    def delete_port_group(self, dvs_id, portgroup_id):
        """Deletes a portgroup."""
        uri = '/api/2.0/xvs/switches/%s/networks/%s' % (dvs_id,
                                                        portgroup_id)
        return self.do_request(HTTP_DELETE, uri, format='xml', decode=False)

    def get_vdn_switch(self, dvs_id):
        uri = '/api/2.0/vdn/switches/%s' % dvs_id
        return self.do_request(HTTP_GET, uri, decode=True)

    def update_vdn_switch(self, switch):
        uri = '/api/2.0/vdn/switches'
        return self.do_request(HTTP_PUT, uri, switch)

    def query_interface(self, edge_id, vnic_index):
        uri = "%s/%s/vnics/%d" % (URI_PREFIX, edge_id, vnic_index)
        return self.do_request(HTTP_GET, uri, decode=True)

    def reconfigure_dhcp_service(self, edge_id, request_config):
        """Reconfigure dhcp static bindings in the created Edge."""
        uri = "/api/4.0/edges/%s/dhcp/config" % edge_id

        return self.do_request(HTTP_PUT, uri, request_config)

    def query_dhcp_configuration(self, edge_id):
        """Query DHCP configuration from the specific edge."""
        uri = "/api/4.0/edges/%s/dhcp/config" % edge_id
        return self.do_request(HTTP_GET, uri)

    def create_dhcp_binding(self, edge_id, request_config):
        """Append one dhcp static binding on the edge."""
        uri = self._build_uri_path(edge_id,
                                   DHCP_SERVICE, DHCP_BINDING_RESOURCE)
        return self.do_request(HTTP_POST, uri, request_config, decode=False)

    def delete_dhcp_binding(self, edge_id, binding_id):
        """Delete one dhcp static binding on the edge."""
        uri = self._build_uri_path(edge_id,
                                   DHCP_SERVICE, DHCP_BINDING_RESOURCE,
                                   binding_id)
        return self.do_request(HTTP_DELETE, uri, decode=False)

    def get_dhcp_binding(self, edge_id, binding_id):
        """Get a dhcp static binding from the edge."""
        uri = self._build_uri_path(edge_id,
                                   DHCP_SERVICE, DHCP_BINDING_RESOURCE,
                                   binding_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    def create_security_group(self, request):
        """Creates a security group container in nsx.

        The method will return the security group ID.
        """
        uri = '%s/globalroot-0' % (SECURITYGROUP_PREFIX)
        return self.do_request(HTTP_POST, uri, request, format='xml',
                               decode=False)

    def delete_security_group(self, securitygroup_id):
        """Deletes a security group container."""
        uri = '%s/%s?force=true' % (SECURITYGROUP_PREFIX, securitygroup_id)
        return self.do_request(HTTP_DELETE, uri, format='xml', decode=False)

    def update_security_group(self, sg_id, sg_name, description):
        """Updates the NSXv security group name."""
        uri = '%s/%s' % (SECURITYGROUP_PREFIX, sg_id)
        h, c = self.do_request(HTTP_GET, uri, format='xml', decode=False)
        sg = et.fromstring(c)
        sg.find('name').text = sg_name
        sg.find('description').text = description
        return self.do_request(HTTP_PUT, uri, et.tostring(sg),
                               format='xml', decode=False, encode=False)

    def list_security_groups(self):
        uri = '%s/scope/globalroot-0' % SECURITYGROUP_PREFIX
        return self.do_request(HTTP_GET, uri, format='xml', decode=False)

    def get_security_group_id(self, sg_name):
        """Returns NSXv security group id which match the given name."""
        h, secgroups = self.list_security_groups()
        root = utils.normalize_xml(secgroups)
        for sg in root.iter('securitygroup'):
            if sg.find('name').text == sg_name:
                return sg.find('objectId').text

    @retry_upon_exception(exceptions.VcnsApiException)
    def create_bridge(self, edge_id, request):
        """Create a bridge."""
        uri = self._build_uri_path(edge_id, BRIDGE)
        return self.do_request(HTTP_PUT, uri, request, format='xml',
                               decode=False)

    @retry_upon_exception(exceptions.VcnsApiException)
    def delete_bridge(self, edge_id):
        """Delete a bridge."""
        uri = self._build_uri_path(edge_id, BRIDGE)
        return self.do_request(HTTP_DELETE, uri, format='xml', decode=False)

    def create_redirect_section(self, request):
        """Creates a layer 3 redirect section in nsx rule table.

        The method will return the uri to newly created section.
        """
        sec_type = FIREWALL_REDIRECT_SEC_TYPE
        uri = '%s/%s?autoSaveDraft=false' % (FIREWALL_PREFIX, sec_type)
        uri += '&operation=insert_before&anchorId=1002'
        return self.do_request(HTTP_POST, uri, request, format='xml',
                               decode=False, encode=False)

    def create_section(self, type, request,
                       insert_top=False, insert_before=None):
        """Creates a layer 3 or layer 2 section in nsx rule table.

        The method will return the uri to newly created section.
        """
        if type == 'ip':
            sec_type = 'layer3sections'
        else:
            sec_type = 'layer2sections'
        uri = '%s/%s?autoSaveDraft=false' % (FIREWALL_PREFIX, sec_type)
        if insert_top:
            uri += '&operation=insert_top'
        # We want to place security-group sections before the default cluster
        # section, and we want to place the default cluster section before the
        # global default section.
        elif insert_before:
            uri += '&operation=insert_before&anchorId=%s' % insert_before
        else:
            uri += '&operation=insert_before&anchorId=1003'
        return self.do_request(HTTP_POST, uri, request, format='xml',
                               decode=False, encode=False)

    def update_section(self, section_uri, request, h):
        """Replaces a section in nsx rule table."""
        uri = '%s?autoSaveDraft=false' % section_uri
        headers = self._get_section_header(section_uri, h)
        return self.do_request(HTTP_PUT, uri, request, format='xml',
                               decode=False, encode=False, headers=headers)

    def delete_section(self, section_uri):
        """Deletes a section in nsx rule table."""
        uri = '%s?autoSaveDraft=false' % section_uri
        return self.do_request(HTTP_DELETE, uri, format='xml', decode=False)

    def get_section(self, section_uri):
        return self.do_request(HTTP_GET, section_uri, format='xml',
                               decode=False)

    def get_default_l3_id(self):
        """Retrieve the id of the default l3 section."""
        h, firewall_config = self.get_dfw_config()
        root = utils.normalize_xml(firewall_config)
        for child in root:
            if str(child.tag) == 'layer3Sections':
                sections = list(child.iter('section'))
                default = sections[-1]
                return default.attrib['id']

    def get_dfw_config(self):
        uri = FIREWALL_PREFIX
        return self.do_request(HTTP_GET, uri, decode=False, format='xml')

    def update_dfw_config(self, request, h):
        uri = FIREWALL_PREFIX
        headers = self._get_section_header(None, h)
        return self.do_request(HTTP_PUT, uri, request, format='xml',
                               decode=False, encode=False, headers=headers)

    def get_section_id(self, section_name):
        """Retrieve the id of a section from nsx."""
        h, firewall_config = self.get_dfw_config()
        root = utils.normalize_xml(firewall_config)
        for sec in root.iter('section'):
            if sec.attrib['name'] == section_name:
                return sec.attrib['id']

    def update_section_by_id(self, id, type, request):
        """Update a section while building its uri from the id."""
        if type == 'ip':
            sec_type = 'layer3sections'
        else:
            sec_type = 'layer2sections'
        section_uri = '%s/%s/%s' % (FIREWALL_PREFIX, sec_type, id)
        self.update_section(section_uri, request, h=None)

    def _get_section_header(self, section_uri, h=None):
        if h is None:
            h, c = self.do_request(HTTP_GET, section_uri, format='xml',
                                   decode=False)
        etag = h['etag']
        headers = {'If-Match': etag}
        return headers

    def remove_rule_from_section(self, section_uri, rule_id):
        """Deletes a rule from nsx section table."""
        uri = '%s/rules/%s?autoSaveDraft=false' % (section_uri, rule_id)
        headers = self._get_section_header(section_uri)
        return self.do_request(HTTP_DELETE, uri, format='xml',
                               headers=headers)

    @retry_upon_exception(exceptions.RequestBad)
    def add_member_to_security_group(self, security_group_id, member_id):
        """Adds a vnic member to nsx security group."""
        uri = '%s/%s/members/%s?failIfExists=false' % (
            SECURITYGROUP_PREFIX, security_group_id, member_id)
        return self.do_request(HTTP_PUT, uri, format='xml', decode=False)

    def remove_member_from_security_group(self, security_group_id,
                                          member_id):
        """Removes a vnic member from nsx security group."""
        uri = '%s/%s/members/%s?failIfAbsent=false' % (
            SECURITYGROUP_PREFIX, security_group_id, member_id)
        return self.do_request(HTTP_DELETE, uri, format='xml', decode=False)

    def set_system_control(self, edge_id, prop):
        uri = self._build_uri_path(edge_id, SYSCTL_SERVICE)

        payload = {
            'featureType': 'systemcontrol',
            'property': prop
        }
        return self.do_request(HTTP_PUT, uri, payload, decode=True)

    def get_system_control(self, edge_id):
        uri = self._build_uri_path(edge_id, SYSCTL_SERVICE)

        return self.do_request(HTTP_GET, uri)

    def _get_enforcement_point_body(self, enforcement_points):
        e_point_list = []
        for enforcement_point in enforcement_points:
            e_point_list.append({
                'enforcementPoint': {
                    'id': enforcement_point,
                    'type': enforcement_point.split('-')[0]
                }
            })
        return {'__enforcementPoints': e_point_list}

    @retry_upon_exception_exclude_error_codes(
        exceptions.RequestBad, [constants.NSX_ERROR_ALREADY_HAS_SG_POLICY])
    def create_spoofguard_policy(self, enforcement_points, name, enable):
        uri = '%s/policies/' % SPOOFGUARD_PREFIX

        body = {'spoofguardPolicy':
                {'name': name,
                 'operationMode': 'MANUAL' if enable else 'DISABLE',
                 'allowLocalIPs': 'true'}}
        body['spoofguardPolicy'].update(
            self._get_enforcement_point_body(enforcement_points))
        return self.do_request(HTTP_POST, uri, body,
                               format='xml', encode=True, decode=False)

    @retry_upon_exception(exceptions.RequestBad)
    def update_spoofguard_policy(self, policy_id,
                                 enforcement_points, name, enable):
        update_uri = '%s/policies/%s' % (SPOOFGUARD_PREFIX, policy_id)
        publish_uri = '%s/%s?action=publish' % (SPOOFGUARD_PREFIX, policy_id)

        body = {'spoofguardPolicy':
                {'policyId': policy_id,
                 'name': name,
                 'operationMode': 'MANUAL' if enable else 'DISABLE',
                 'allowLocalIPs': 'true'}}
        body['spoofguardPolicy'].update(
            self._get_enforcement_point_body(enforcement_points))

        self.do_request(HTTP_PUT, update_uri, body,
                        format='xml', encode=True, decode=False)
        return self.do_request(HTTP_POST, publish_uri, decode=False)

    @retry_upon_exception(exceptions.RequestBad)
    def delete_spoofguard_policy(self, policy_id):
        uri = '%s/policies/%s' % (SPOOFGUARD_PREFIX, policy_id)
        return self.do_request(HTTP_DELETE, uri, decode=False)

    def get_spoofguard_policy(self, policy_id):
        uri = '%s/policies/%s' % (SPOOFGUARD_PREFIX, policy_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    def get_spoofguard_policies(self):
        uri = '%s/policies/' % SPOOFGUARD_PREFIX
        return self.do_request(HTTP_GET, uri, decode=True)

    def _approve_assigned_addresses(self, policy_id,
                                    vnic_id, mac_addr, addresses):
        uri = '%s/%s' % (SPOOFGUARD_PREFIX, policy_id)
        addresses = [{'ipAddress': ip_addr} for ip_addr in addresses]
        body = {'spoofguardList':
                {'spoofguard':
                 {'id': vnic_id,
                  'vnicUuid': vnic_id,
                  'approvedIpAddress': addresses,
                  'approvedMacAddress': mac_addr,
                  'publishedIpAddress': addresses,
                  'publishedMacAddress': mac_addr}}}
        try:
            return self.do_request(HTTP_POST, '%s?action=approve' % uri,
                                   body, format='xml', decode=False)
        except exceptions.VcnsApiException as e:
            nsx_errcode = self.xmlapi_client._get_nsx_errorcode(e.response)
            if nsx_errcode == constants.NSX_ERROR_ALREADY_EXISTS:
                LOG.warning("Spoofguard entry for %s already exists",
                            vnic_id)
                raise exceptions.AlreadyExists(resource=vnic_id)
            # raise original exception for retries
            raise

    @retry_upon_exception(exceptions.RequestBad)
    def approve_assigned_addresses(self, policy_id,
                                   vnic_id, mac_addr, addresses):
        return self._approve_assigned_addresses(
            policy_id, vnic_id, mac_addr, addresses)

    @retry_upon_exception(exceptions.VcnsApiException)
    def publish_assigned_addresses(self, policy_id, vnic_id):
        uri = '%s/%s' % (SPOOFGUARD_PREFIX, policy_id)
        publish_vnic_uri = '%s?action=publish&vnicId=%s' % (uri, vnic_id)
        return self.do_request(HTTP_POST, publish_vnic_uri, decode=False)

    def inactivate_vnic_assigned_addresses(self, policy_id, vnic_id):
        try:
            self._approve_assigned_addresses(policy_id, vnic_id, '', [])
        except exceptions.RequestBad:
            LOG.debug("Request failed: inactivate vnic %s assigned addresses",
                      vnic_id)
        else:
            return self.publish_assigned_addresses(policy_id, vnic_id)

    def _build_uri_path(self, edge_id,
                        service,
                        resource=None,
                        resource_id=None,
                        parent_resource_id=None,
                        fields=None,
                        relations=None,
                        filters=None,
                        types=None,
                        is_attachment=False,
                        is_async=False):
        uri_prefix = "%s/%s/%s" % (URI_PREFIX, edge_id, service)
        if resource:
            res_path = resource + (resource_id and "/%s" % resource_id or '')
            uri_path = "%s/%s" % (uri_prefix, res_path)
        else:
            uri_path = uri_prefix
        if is_async:
            return (uri_path + "?async=true")
        else:
            return uri_path

    def add_vm_to_exclude_list(self, vm_id):
        uri = '%s/%s' % (EXCLUDELIST_PREFIX, vm_id)
        return self.do_request(HTTP_PUT, uri)

    def delete_vm_from_exclude_list(self, vm_id):
        uri = '%s/%s' % (EXCLUDELIST_PREFIX, vm_id)
        return self.do_request(HTTP_DELETE, uri)

    def get_scoping_objects(self):
        uri = '%s/usermgmt/scopingobjects' % SERVICES_PREFIX
        h, scoping_objects = self.do_request(HTTP_GET, uri, decode=False,
                                             format='xml')
        return scoping_objects

    def _scopingobjects_lookup(self, type_names, object_id, name=None,
                               use_cache=False):
        """Look for a specific object in the NSX scoping objects."""
        # used cached scoping objects during plugin init since it is
        # a big structure to retrieve and parse each time.
        if use_cache and self._normalized_scoping_objects is not None:
            # Use the cached data
            root = self._normalized_scoping_objects
        else:
            # Not using cache, or we do want to use it,
            # but it was not saved yet:
            # So get the data from the NSX and parse it
            so_list = self.get_scoping_objects()
            root = utils.normalize_xml(so_list)
            # Save it for possible usage next time (even if not using cache)
            self._normalized_scoping_objects = root

        for obj in root.iter('object'):
            if (obj.find('objectTypeName').text in type_names and
                    obj.find('objectId').text == object_id and
                    (name is None or obj.find('name').text == name)):
                return True

        return False

    def validate_datacenter_moid(self, object_id, during_init=False):
        return self._scopingobjects_lookup(['Datacenter'], object_id,
                                           use_cache=during_init)

    def validate_network(self, object_id, during_init=False):
        return self._scopingobjects_lookup(NETWORK_TYPES, object_id,
                                           use_cache=during_init)

    def validate_network_name(self, object_id, name, during_init=False):
        return self._scopingobjects_lookup(NETWORK_TYPES, object_id,
                                           name=name, use_cache=during_init)

    def validate_vdn_scope(self, object_id):
        uri = '%s/scopes' % VDN_PREFIX
        h, scope_list = self.do_request(HTTP_GET, uri, decode=False,
                                        format='xml')
        root = utils.normalize_xml(scope_list)
        for obj_id in root.iter('objectId'):
            if obj_id.text == object_id:
                return True

        return False

    def get_dvs_list(self):
        uri = '%s/switches' % VDN_PREFIX
        h, dvs_list = self.do_request(HTTP_GET, uri, decode=False,
                                      format='xml')
        root = utils.normalize_xml(dvs_list)
        dvs_list = []
        for obj_id in root.iter('objectId'):
            if obj_id.text:
                dvs_list.append(obj_id.text)

        return dvs_list

    def validate_dvs(self, object_id, dvs_list=None):
        if not dvs_list:
            dvs_list = self.get_dvs_list()
        for dvs in dvs_list:
            if dvs == object_id:
                return True

        return False

    def validate_inventory(self, object_id):
        uri = '%s/inventory/%s/basicinfo' % (SERVICES_PREFIX, object_id)
        try:
            h, c = self.do_request(HTTP_GET, uri, decode=False)
        except exceptions.ResourceNotFound:
            return False
        return True

    def get_inventory_name(self, object_id):
        uri = '%s/inventory/%s/basicinfo' % (SERVICES_PREFIX, object_id)
        h, c = self.do_request(HTTP_GET, uri, decode=True)
        return c['name']

    def _get_version(self):
        uri = '/api/2.0/services/vsmconfig'
        h, c = self.do_request(HTTP_GET, uri, decode=True)
        version = c['version']
        LOG.debug("NSX Version: %s", version)
        return version

    def get_version(self):
        if self._nsx_version is None:
            try:
                self._nsx_version = self._get_version()
            except Exception as e:
                # Versions prior to 6.2.0 do not support the above API
                LOG.error("Unable to get NSX version. Exception: %s", e)
                # Minimum supported version is 6.1
                self._nsx_version = '6.1'
        return self._nsx_version

    def get_tuning_configration(self):
        uri = '/api/4.0/edgePublish/tuningConfiguration'
        h, c = self.do_request(HTTP_GET, uri, decode=True)
        return c

    def configure_aggregate_publishing(self):
        uri = "/api/4.0/edgePublish/tuningConfiguration"
        # Ensure that configured values are not changed
        config = self.get_tuning_configration()
        LOG.debug("Tuning configuration: %s", config)
        tuning = et.Element('tuningConfiguration')
        for opt, val in six.iteritems(config):
            child = et.Element(opt)
            if opt == 'aggregatePublishing':
                child.text = 'true'
            else:
                child.text = str(val)
            tuning.append(child)
        return self.do_request(HTTP_PUT, uri, et.tostring(tuning),
                               format='xml', decode=True)

    def configure_reservations(self):
        uri = "/api/4.0/edgePublish/tuningConfiguration"
        config = self.get_tuning_configration()
        tuning = et.Element('tuningConfiguration')
        for opt, val in six.iteritems(config):
            child = et.Element(opt)
            if (opt == 'edgeVCpuReservationPercentage' or
                opt == 'edgeMemoryReservationPercentage'):
                child.text = '0'
            elif opt == 'megaHertzPerVCpu':
                child.text = '1500'
            else:
                child.text = str(val)
            tuning.append(child)
        return self.do_request(HTTP_PUT, uri, et.tostring(tuning),
                               format='xml', decode=True)

    def enable_ha(self, edge_id, request_config):
        """Enable HA in the given edge."""
        uri = "/api/4.0/edges/%s/highavailability/config" % edge_id
        return self.do_request(HTTP_PUT, uri, request_config)

    def change_edge_appliance_size(self, edge_id, size):
        """Change the size of edge appliances."""
        uri = ("/api/4.0/edges/%s/appliances/?size=%s" %
               (edge_id, size))
        return self.do_request(HTTP_POST, uri)

    def change_edge_appliance(self, edge_id, request):
        uri = "/api/4.0/edges/%s/appliances" % edge_id
        return self.do_request(HTTP_PUT, uri, request)

    def get_edge_appliances(self, edge_id):
        uri = "/api/4.0/edges/%s/appliances" % edge_id
        return self.do_request(HTTP_GET, uri)

    def upload_edge_certificate(self, edge_id, request):
        """Creates a certificate on the specified Edge appliance."""
        uri = '%s/%s/%s' % (TRUSTSTORE_PREFIX, CERTIFICATE, edge_id)
        return self.do_request(HTTP_POST, uri, request, decode=True)

    def create_csr(self, edge_id, request=nsxv_constants.CSR_REQUEST):
        """Create a CSR on the specified Edge appliance."""
        uri = '%s/%s/%s' % (TRUSTSTORE_PREFIX, CSR, edge_id)
        return self.do_request(HTTP_POST, uri, request, format='xml',
                               decode=False)

    def create_csr_cert(self, csr_id):
        """Create a CSR self signed cert on the specified Edge appliance."""
        uri = '%s/%s/%s?noOfDays=%s' % (TRUSTSTORE_PREFIX, CSR, csr_id,
                                        nsxv_constants.CERT_NUMBER_OF_DAYS)
        return self.do_request(HTTP_PUT, uri)

    def get_service_insertion_profile(self, profile_id):
        profiles_uri = '%s/%s' % (SERVICE_INSERTION_PROFILE_PREFIX, profile_id)
        return self.do_request(HTTP_GET, profiles_uri, format='xml',
                               decode=False)

    def update_service_insertion_profile_binding(self, profile_id, request):
        profiles_uri = '%s/%s/%s' % (SERVICE_INSERTION_PROFILE_PREFIX,
                                     profile_id, 'binding')
        return self.do_request(HTTP_POST, profiles_uri, request, format='xml',
                               decode=False)

    def create_ipam_ip_pool(self, request):
        uri = '%s/%s/%s' % (SERVICES_PREFIX, IPAM_POOL_SERVICE,
                            IPAM_POOL_SCOPE)
        return self.do_request(HTTP_POST, uri, request, format='xml',
                               decode=False)

    def delete_ipam_ip_pool(self, pool_id):
        uri = '%s/%s/%s' % (SERVICES_PREFIX, IPAM_POOL_SERVICE, pool_id)
        return self.do_request(HTTP_DELETE, uri)

    def get_ipam_ip_pool(self, pool_id):
        uri = '%s/%s/%s' % (SERVICES_PREFIX, IPAM_POOL_SERVICE, pool_id)
        return self.do_request(HTTP_GET, uri, decode=True)

    def allocate_ipam_ip_from_pool(self, pool_id, ip_addr=None):
        uri = '%s/%s/%s/%s' % (SERVICES_PREFIX, IPAM_POOL_SERVICE, pool_id,
                               'ipaddresses')
        if ip_addr:
            request = {'ipAddressRequest': {'allocationMode': 'RESERVE',
                                            'ipAddress': ip_addr}}
        else:
            request = {'ipAddressRequest': {'allocationMode': 'ALLOCATE'}}
        return self.do_request(HTTP_POST, uri, request, format='xml',
                               decode=False)

    def release_ipam_ip_to_pool(self, pool_id, ip_addr):
        uri = '%s/%s/%s/%s/%s' % (SERVICES_PREFIX, IPAM_POOL_SERVICE, pool_id,
                               'ipaddresses', ip_addr)
        return self.do_request(HTTP_DELETE, uri)

    def get_security_policy(self, policy_id, return_xml=True):
        # get the policy configuration as an xml string / dictionary
        uri = '%s/%s' % (SECURITY_POLICY_PREFIX, policy_id)
        if return_xml:
            format = 'xml'
            decode = False
        else:
            format = 'json'
            decode = True
        h, policy = self.do_request(HTTP_GET, uri, format=format,
                                    decode=decode)
        return policy

    def update_security_policy(self, policy_id, request):
        # update the policy configuration. request should be an xml string
        uri = '%s/%s' % (SECURITY_POLICY_PREFIX, policy_id)
        return self.do_request(HTTP_PUT, uri, request,
                               format='xml',
                               decode=False, encode=True)

    def get_security_policies(self):
        # get the policies configuration dictionary
        uri = '%s/all' % (SECURITY_POLICY_PREFIX)
        h, policies = self.do_request(HTTP_GET, uri, decode=True)
        return policies

    def list_applications(self):
        uri = '%s/scope/globalroot-0' % APPLICATION_PREFIX
        h, apps = self.do_request(HTTP_GET, uri, decode=True)
        return apps

    def update_edge_routing_config(self, edge_id, request_config):
        uri = self._build_uri_path(edge_id, ROUTING_CONFIG)
        return self.do_request(HTTP_PUT, uri,
                               VcnsApiClient.xmldumps(request_config),
                               format='xml')

    def get_edge_routing_config(self, edge_id):
        uri = self._build_uri_path(edge_id, ROUTING_CONFIG)
        return self.do_request(HTTP_GET, uri)

    def update_bgp_dynamic_routing(self, edge_id, bgp_request):
        uri = self._build_uri_path(edge_id, BGP_ROUTING_CONFIG)
        return self.do_request(HTTP_PUT, uri,
                               VcnsApiClient.xmldumps(bgp_request),
                               format='xml')

    def get_bgp_routing_config(self, edge_id):
        uri = self._build_uri_path(edge_id, BGP_ROUTING_CONFIG)
        return self.do_request(HTTP_GET, uri)

    def delete_bgp_routing_config(self, edge_id):
        uri = self._build_uri_path(edge_id, BGP_ROUTING_CONFIG)
        return self.do_request(HTTP_DELETE, uri)

    def get_global_objects(self):
        uri = '%s/application/scope/globalroot-0' % SERVICES_PREFIX
        h, scoping_objects = self.do_request(HTTP_GET, uri, decode=False,
                                             format='xml')
        return scoping_objects

    def _globalobjects_lookup(self, name, use_cache=False):
        """Return objectId a specific name in the NSX global objects."""
        # used cached scoping objects during plugin init since it is
        # a big structure to retrieve and parse each time.
        if use_cache and self._normalized_global_objects is not None:
            # Use the cached data
            root = self._normalized_global_objects
        else:
            # Not using cache, or we do want to use it,
            # but it was not saved yet:
            # So get the data from the NSX and parse it
            so_list = self.get_global_objects()
            root = utils.normalize_xml(so_list)
            # Save it for possible usage next time (even if not using cache)
            self._normalized_global_objects = root

        for obj in root.iter('application'):
            if obj.find('name').text == name:
                return obj.find('objectId').text

    def get_application_id(self, name):
        return self._globalobjects_lookup(name, use_cache=True)



# Copyright 2013 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import re

from distutils import version
import functools
import hashlib
import xml.etree.ElementTree as et

import eventlet
import six
import tenacity
from tenacity import _utils as tenacity_utils

from neutron import version as n_version
from neutron_lib.api import validators
from neutron_lib import constants
from oslo_context import context as common_context
from oslo_log import log

from vmware_nsxlib.v3 import nsx_constants as v3_const

LOG = log.getLogger(__name__)

MAX_DISPLAY_NAME_LEN = 40
NEUTRON_VERSION = n_version.version_info.release_string()
OS_NEUTRON_ID_SCOPE = 'os-neutron-id'


# Allowed network types for the NSX Plugin
class NetworkTypes(object):
    """Allowed provider network types for the NSX Plugin."""
    L3_EXT = 'l3_ext'
    STT = 'stt'
    GRE = 'gre'
    FLAT = 'flat'
    VLAN = 'vlan'
    BRIDGE = 'bridge'
    PORTGROUP = 'portgroup'


# Allowed network types for the NSX-v Plugin
class NsxVNetworkTypes(object):
    """Allowed provider network types for the NSX-v Plugin."""
    FLAT = 'flat'
    VLAN = 'vlan'
    VXLAN = 'vxlan'
    PORTGROUP = 'portgroup'


# Allowed network types for the NSXv3 Plugin
class NsxV3NetworkTypes(object):
    """Allowed provider network types for the NSXv3 Plugin."""
    FLAT = 'flat'
    VLAN = 'vlan'
    GENEVE = 'geneve'
    NSX_NETWORK = 'nsx-net'


def is_nsx_version_1_1_0(nsx_version):
    return (version.LooseVersion(nsx_version) >=
            version.LooseVersion(v3_const.NSX_VERSION_1_1_0))


def is_nsx_version_2_0_0(nsx_version):
    return (version.LooseVersion(nsx_version) >=
            version.LooseVersion(v3_const.NSX_VERSION_2_0_0))


def is_nsx_version_2_1_0(nsx_version):
    return (version.LooseVersion(nsx_version) >=
            version.LooseVersion(v3_const.NSX_VERSION_2_1_0))


def is_nsxv_version_6_2(nsx_version):
    return (version.LooseVersion(nsx_version) >=
            version.LooseVersion('6.2'))


def is_nsxv_version_6_3(nsx_version):
    return (version.LooseVersion(nsx_version) >=
            version.LooseVersion('6.3'))


def is_nsxv_dhcp_binding_supported(nsx_version):
    return ((version.LooseVersion(nsx_version) >=
             version.LooseVersion('6.3.3')) or
            (version.LooseVersion(nsx_version) >=
             version.LooseVersion('6.2.8') and
             version.LooseVersion(nsx_version) <
             version.LooseVersion('6.3')))


def get_tags(**kwargs):
    tags = ([dict(tag=value, scope=key)
            for key, value in six.iteritems(kwargs)])
    tags.append({"tag": NEUTRON_VERSION, "scope": "quantum"})
    return sorted(tags, key=lambda x: x['tag'])


def device_id_to_vm_id(device_id, obfuscate=False):
    # device_id can be longer than 40 characters, for example
    # a device_id for a dhcp port is like the following:
    #
    # dhcp83b5fdeb-e3b4-5e18-ac5f-55161...80747326-47d7-46c2-a87a-cf6d5194877c
    #
    # To fit it into an NSX tag we need to hash it, however device_id
    # used for ports associated to VM's are small enough so let's skip the
    # hashing
    if len(device_id) > MAX_DISPLAY_NAME_LEN or obfuscate:
        return hashlib.sha1(device_id.encode()).hexdigest()
    else:
        return device_id or "N/A"


def check_and_truncate(display_name):
    if (validators.is_attr_set(display_name) and
            len(display_name) > MAX_DISPLAY_NAME_LEN):
        LOG.debug("Specified name:'%s' exceeds maximum length. "
                  "It will be truncated on NSX", display_name)
        return display_name[:MAX_DISPLAY_NAME_LEN]
    return display_name or ''


def normalize_xml(data):
    data = data.encode('ascii', 'ignore')
    return et.fromstring(data)


def _get_bad_request_error_code(e):
    """Get the error code out of the exception"""
    try:
        desc = normalize_xml(e.response)
        return int(desc.find('errorCode').text)
    except Exception:
        pass


def _log_before_retry(func, trial_number):
    """Before call strategy that logs to some logger the attempt."""
    if trial_number > 1:
        LOG.warning("Retrying call to '%(func)s' for the %(num)s time",
                    {'func': tenacity_utils.get_callback_name(func),
                     'num': tenacity_utils.to_ordinal(trial_number)})


def _get_args_from_frame(frames, frame_num):
    if len(frames) > frame_num and frames[frame_num] and frames[frame_num][0]:
        argvalues = inspect.getargvalues(frames[frame_num][0])
        formated_args = inspect.formatargvalues(*argvalues)
        # remove the first 'self' arg from the log as it adds no information
        formated_args = re.sub(r'\(self=.*?, ', "(", formated_args)
        return formated_args


def _log_after_retry(func, trial_number, trial_time_taken):
    """After call strategy that logs to some logger the finished attempt."""
    # Using inspect to get arguments of the relevant call
    frames = inspect.trace()
    formated_args = _get_args_from_frame(frames, 1)
    if not formated_args:
        formated_args = "Unknown"

    LOG.warning("Finished retry of %(func)s for the %(num)s time after "
                "%(time)0.3f(s) with args: %(args)s",
                {'func': tenacity_utils.get_callback_name(func),
                 'num': tenacity_utils.to_ordinal(trial_number),
                 'time': trial_time_taken,
                 'args': formated_args})


def retry_upon_exception_exclude_error_codes(
    exc, excluded_errors, delay, max_delay, max_attempts):
    """Retry with the configured exponential delay, unless the exception error
    code is in the given list
    """
    def retry_if_not_error_codes(e):
        # return True only for BadRequests without error codes or with error
        # codes not in the exclude list
        if isinstance(e, exc):
            error_code = _get_bad_request_error_code(e)
            if error_code and error_code not in excluded_errors:
                return True
        return False

    return tenacity.retry(reraise=True,
                          retry=tenacity.retry_if_exception(
                                retry_if_not_error_codes),
                          wait=tenacity.wait_exponential(
                                multiplier=delay, max=max_delay),
                          stop=tenacity.stop_after_attempt(max_attempts),
                          before=_log_before_retry, after=_log_after_retry)


def retry_upon_exception(exc, delay, max_delay, max_attempts):
    return tenacity.retry(reraise=True,
                          retry=tenacity.retry_if_exception_type(exc),
                          wait=tenacity.wait_exponential(
                                multiplier=delay, max=max_delay),
                          stop=tenacity.stop_after_attempt(max_attempts),
                          before=_log_before_retry, after=_log_after_retry)


def read_file(path):
    try:
        with open(path) as file:
            return file.read().strip()
    except IOError as e:
        LOG.error("Error while opening file "
                  "%(path)s: %(err)s", {'path': path, 'err': str(e)})


def get_name_and_uuid(name, uuid, tag=None, maxlen=80):
    short_uuid = '_' + uuid[:5] + '...' + uuid[-5:]
    maxlen = maxlen - len(short_uuid)
    if tag:
        maxlen = maxlen - len(tag) - 1
        return name[:maxlen] + '_' + tag + short_uuid
    else:
        return name[:maxlen] + short_uuid


def is_ipv4_ip_address(addr):

    def _valid_part(part):
        try:
            int_part = int(part)
            if int_part < 0 or int_part > 255:
                return False
            return True
        except ValueError:
            return False

    parts = str(addr).split('.')
    if len(parts) != 4:
        return False

    for ip_part in parts:
        if not _valid_part(ip_part):
            return False
    return True


def is_port_dhcp_configurable(port):
    owner = port.get('device_owner')
    return (owner and
            not owner.startswith(constants.DEVICE_OWNER_NETWORK_PREFIX))


def spawn_n(func, *args, **kwargs):
    """Passthrough method for eventlet.spawn_n.

    This utility exists so that it can be stubbed for testing without
    interfering with the service spawns.

    It will also grab the context from the threadlocal store and add it to
    the store on the new thread.  This allows for continuity in logging the
    context when using this method to spawn a new thread.
    """
    _context = common_context.get_current()

    @functools.wraps(func)
    def context_wrapper(*args, **kwargs):
        # NOTE: If update_store is not called after spawn_n it won't be
        # available for the logger to pull from threadlocal storage.
        if _context is not None:
            _context.update_store()
        func(*args, **kwargs)

    eventlet.spawn_n(context_wrapper, *args, **kwargs)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import log
import six

LOG = log.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class BaseJob(object):

    _core_plugin = None

    def __init__(self, readonly):
        self.readonly = readonly or (self.get_name() in
                                     cfg.CONF.nsxv.housekeeping_readonly_jobs)
        LOG.info('Housekeeping: %s job initialized in %s mode',
                 self.get_name(), 'RO' if self.readonly else 'RW')

    @property
    def plugin(self):
        if not self._core_plugin:
            self._core_plugin = directory.get_plugin()
            if self._core_plugin.is_tvd_plugin() is True:
                # get the plugin that match this driver
                self._core_plugin = self.get_project_plugin(
                    self._core_plugin)
        return self._core_plugin

    @abc.abstractmethod
    def get_name(self):
        pass

    @abc.abstractmethod
    def get_description(self):
        pass

    @abc.abstractmethod
    def run(self, context):
        pass

    @abc.abstractmethod
    def get_project_plugin(self, plugin):
        pass



# Copyright 2014 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from neutron.db import l3_db
from neutron_lib import constants as const

# A unique MAC to quickly identify the LSN port used for metadata services
# when dhcp on the subnet is off. Inspired by leet-speak for 'metadata'.
METADATA_MAC = "fa:15:73:74:d4:74"
METADATA_PORT_ID = 'metadata:id'
METADATA_PORT_NAME = 'metadata:name'
METADATA_DEVICE_ID = 'metadata:device'
SPECIAL_OWNERS = (const.DEVICE_OWNER_DHCP,
                  const.DEVICE_OWNER_ROUTER_GW,
                  l3_db.DEVICE_OWNER_ROUTER_INTF)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import l3_db
from neutron.services.flavors import flavors_plugin
from neutron_lib import exceptions as n_exc

from vmware_nsx._i18n import _
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas import lb_const
from vmware_nsxlib.v3 import utils


def get_tags(plugin, resource_id, resource_type, project_id, project_name):
    resource = {'project_id': project_id,
                'id': resource_id}
    tags = plugin.nsxlib.build_v3_tags_payload(
        resource, resource_type=resource_type,
        project_name=project_name)
    return tags


def get_network_from_subnet(context, plugin, subnet_id):
    subnet = plugin.get_subnet(context, subnet_id)
    if subnet:
        return plugin.get_network(context, subnet['network_id'])


def get_router_from_network(context, plugin, subnet_id):
    subnet = plugin.get_subnet(context, subnet_id)
    network_id = subnet['network_id']
    port_filters = {'device_owner': [l3_db.DEVICE_OWNER_ROUTER_INTF],
                    'network_id': [network_id]}
    ports = plugin.get_ports(context, filters=port_filters)
    if ports:
        router = plugin.get_router(context, ports[0]['device_id'])
        if router.get('external_gateway_info'):
            return router['id']


def get_lb_router_id(context, plugin, lb):
    router_client = plugin.nsxlib.logical_router
    name = utils.get_name_and_uuid(lb.name or 'router', lb.id)
    tags = get_tags(plugin, lb.id, lb_const.LB_LB_TYPE, lb.tenant_id,
                    context.project_name)
    edge_cluster_uuid = plugin._get_edge_cluster(plugin._default_tier0_router)
    lb_router = router_client.create(name, tags, edge_cluster_uuid)
    return lb_router


def get_lb_flavor_size(flavor_plugin, context, flavor_id):
    if not flavor_id:
        return lb_const.DEFAULT_LB_SIZE
    else:
        flavor = flavors_plugin.FlavorsPlugin.get_flavor(
            flavor_plugin, context, flavor_id)
        flavor_size = flavor['name']
        if flavor_size in lb_const.LB_FLAVOR_SIZES:
            return flavor_size.upper()
        else:
            err_msg = (_("Invalid flavor size %(flavor)s, only 'small', "
                         "'medium', or 'large' are supported") %
                       {'flavor': flavor_size})
            raise n_exc.InvalidInput(error_message=err_msg)


def validate_lb_subnet(context, plugin, subnet_id):
    '''Validate LB subnet before creating loadbalancer on it.

    To create a loadbalancer, the network has to be either an external
    network or private network that connects to a tenant router. The
    tenant router needs to connect to gateway. It will throw
    exception if the network doesn't meet this requirement.

    :param context: context
    :param plugin: core plugin
    :param subnet_id: loadbalancer's subnet id
    :return: True if subnet meet requirement, otherwise return False
    '''
    network = get_network_from_subnet(context, plugin, subnet_id)
    valid_router = get_router_from_network(
        context, plugin, subnet_id)
    if network.get('router:external') or valid_router:
        return True
    else:
        return False


def get_rule_match_conditions(policy):
    match_conditions = []
    # values in rule have already been validated in LBaaS API,
    # we won't need to valid anymore in driver, and just get
    # the LB rule mapping from the dict.
    for rule in policy.rules:
        match_type = lb_const.LB_RULE_MATCH_TYPE[rule.compare_type]
        if rule.type == lb_const.L7_RULE_TYPE_COOKIE:
            header_value = rule.key + '=' + rule.value
            match_conditions.append(
                {'type': 'LbHttpRequestHeaderCondition',
                 'match_type': match_type,
                 'header_name': 'Cookie',
                 'header_value': header_value})
        elif rule.type == lb_const.L7_RULE_TYPE_FILE_TYPE:
            match_conditions.append(
                {'type': 'LbHttpRequestUriCondition',
                 'match_type': match_type,
                 'uri': '*.' + rule.value})
        elif rule.type == lb_const.L7_RULE_TYPE_HEADER:
            match_conditions.append(
                {'type': 'LbHttpRequestHeaderCondition',
                 'match_type': match_type,
                 'header_name': rule.key,
                 'header_value': rule.value})
        elif rule.type == lb_const.L7_RULE_TYPE_HOST_NAME:
            match_conditions.append(
                {'type': 'LbHttpRequestHeaderCondition',
                 'match_type': match_type,
                 'header_name': 'Host',
                 'header_value': rule.value})
        elif rule.type == lb_const.L7_RULE_TYPE_PATH:
            match_conditions.append(
                {'type': 'LbHttpRequestUriCondition',
                 'match_type': match_type,
                 'uri': rule.value})
        else:
            msg = (_('l7rule type %(type)s is not supported in LBaaS') %
                   {'type': rule.type})
            raise n_exc.BadRequest(resource='lbaas-l7rule', msg=msg)
    return match_conditions


def get_rule_actions(context, l7policy):
    lb_id = l7policy.listener.loadbalancer_id
    if l7policy.action == lb_const.L7_POLICY_ACTION_REDIRECT_TO_POOL:
        pool_binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, lb_id, l7policy.redirect_pool_id)
        if pool_binding:
            lb_pool_id = pool_binding['lb_pool_id']
            actions = [{'type': lb_const.LB_SELECT_POOL_ACTION,
                        'pool_id': lb_pool_id}]
        else:
            msg = _('Failed to get LB pool binding from nsx db')
            raise n_exc.BadRequest(resource='lbaas-l7rule-create',
                                   msg=msg)
    elif l7policy.action == lb_const.L7_POLICY_ACTION_REDIRECT_TO_URL:
        actions = [{'type': lb_const.LB_HTTP_REDIRECT_ACTION,
                    'redirect_status': lb_const.LB_HTTP_REDIRECT_STATUS,
                    'redirect_url': l7policy.redirect_url}]
    elif l7policy.action == lb_const.L7_POLICY_ACTION_REJECT:
        actions = [{'type': lb_const.LB_REJECT_ACTION,
                    'reply_status': lb_const.LB_HTTP_REJECT_STATUS}]
    else:
        msg = (_('Invalid l7policy action: %(action)s') %
               {'action': l7policy.action})
        raise n_exc.BadRequest(resource='lbaas-l7rule-create',
                               msg=msg)
    return actions


def convert_l7policy_to_lb_rule(context, policy):
    return {
        'match_conditions': get_rule_match_conditions(policy),
        'actions': get_rule_actions(context, policy),
        'phase': lb_const.LB_RULE_HTTP_FORWARDING,
        'match_strategy': 'ALL'
    }


def remove_rule_from_policy(rule):
    l7rules = rule.policy.rules
    rule.policy.rules = [r for r in l7rules if r.id != rule.id]



# Copyright 2016 VMware, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Port action implementations with nsx extensions"""

from openstackclient.network.v2 import port
from osc_lib import utils as osc_utils

from vmware_nsx._i18n import _
from vmware_nsx.osc.v2 import utils


def add_nsx_extensions_to_parser(parser, client_manager, is_create=True):
    allowed_extensions = utils.get_extensions(client_manager)
    # Provider security group (only for create action)
    if (is_create and
        'provider-security-group' in allowed_extensions):
        parser.add_argument(
            '--provider-security-group',
            metavar='<provider-security-group>',
            action='append',
            dest='provider_security_groups',
            help=_("Provider Security group to associate with this port "
                   "(name or ID) "
                   "(repeat option to set multiple security groups)")
        )
    if 'vnic-index' in allowed_extensions:
        # vnic index
        parser.add_argument(
            '--vnic-index',
            type=int,
            metavar='<vnic-index>',
            help=_("Vnic index")
        )
    if 'mac-learning' in allowed_extensions:
        # mac-learning-enabled
        mac_learning_group = parser.add_mutually_exclusive_group()
        mac_learning_group.add_argument(
            '--enable-mac-learning',
            action='store_true',
            help=_("Enable MAC learning")
        )
        mac_learning_group.add_argument(
            '--disable-mac-learning',
            action='store_true',
            help=_("Disable MAC learning (Default")
        )


# overriding the port module global method, to add the nsx extensions
super_get_attrs = port._get_attrs


def _get_plugin_attrs(client_manager, parsed_args):
    allowed_extensions = utils.get_extensions(client_manager)
    attrs = super_get_attrs(client_manager, parsed_args)
    # Provider security groups
    if 'provider-security-group' in allowed_extensions:
        if (hasattr(parsed_args, 'provider_security_groups') and
            parsed_args.provider_security_groups is not None):
            attrs['provider_security_groups'] = [
                client_manager.network.find_security_group(
                    sg, ignore_missing=False).id
                for sg in parsed_args.provider_security_groups]

    if 'vnic-index' in allowed_extensions:
        # Vnic index
        if parsed_args.vnic_index is not None:
            attrs['vnic_index'] = parsed_args.vnic_index
            parsed_args.vnic_index = None
    if 'mac-learning' in allowed_extensions:
        # mac-learning-enabled
        if parsed_args.enable_mac_learning:
            attrs['mac_learning_enabled'] = True
        if parsed_args.disable_mac_learning:
            attrs['mac_learning_enabled'] = False

    return attrs


port._get_attrs = _get_plugin_attrs


# Update the port module global _formatters, to format provider security
# groups too
port._formatters['provider_security_groups'] = osc_utils.format_list


class NsxCreatePort(port.CreatePort):
    """Create a new port with vmware nsx extensions """

    def get_parser(self, prog_name):
        # Add the nsx attributes to the neutron port attributes
        parser = super(NsxCreatePort, self).get_parser(prog_name)
        add_nsx_extensions_to_parser(parser, self.app.client_manager,
                                     is_create=True)
        return parser


class NsxSetPort(port.SetPort):
    """Set port properties with vmware nsx extensions """

    def get_parser(self, prog_name):
        # Add the nsx attributes to the neutron port attributes
        parser = super(NsxSetPort, self).get_parser(prog_name)
        add_nsx_extensions_to_parser(parser, self.app.client_manager,
                                     is_create=False)
        return parser



# Copyright 2017 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from neutron_lib.db import constants as db_constants
from neutron_lib.services.qos import base
from neutron_lib.services.qos import constants as qos_consts
from oslo_log import log as logging

from vmware_nsx.services.qos.nsx_v3 import utils as qos_utils

LOG = logging.getLogger(__name__)

DRIVER = None

SUPPORTED_RULES = {
    qos_consts.RULE_TYPE_BANDWIDTH_LIMIT: {
        qos_consts.MAX_KBPS: {
            'type:range': [0, db_constants.DB_INTEGER_MAX_VALUE]},
        qos_consts.MAX_BURST: {
            'type:range': [0, db_constants.DB_INTEGER_MAX_VALUE]},
        qos_consts.DIRECTION: {
            'type:values': [constants.EGRESS_DIRECTION,
                            constants.INGRESS_DIRECTION]}
    },
    qos_consts.RULE_TYPE_DSCP_MARKING: {
        qos_consts.DSCP_MARK: {'type:values': constants.VALID_DSCP_MARKS}
    }
}


class NSXv3QosDriver(base.DriverBase):

    @staticmethod
    def create():
        return NSXv3QosDriver(
            name='NSXv3QosDriver',
            vif_types=None,
            vnic_types=None,
            supported_rules=SUPPORTED_RULES,
            requires_rpc_notifications=False)

    def __init__(self, **kwargs):
        self.handler = qos_utils.QosNotificationsHandler()
        super(NSXv3QosDriver, self).__init__(**kwargs)

    def is_vif_type_compatible(self, vif_type):
        return True

    def is_vnic_compatible(self, vnic_type):
        return True

    def create_policy(self, context, policy):
        self.handler.create_policy(context, policy)

    def update_policy(self, context, policy):
        if (hasattr(policy, "rules")):
            self.handler.update_policy_rules(
                context, policy.id, policy["rules"])

        # May also need to update name / description
        self.handler.update_policy(context, policy.id, policy)

    def delete_policy(self, context, policy):
        self.handler.delete_policy(context, policy.id)

    def update_policy_precommit(self, context, policy):
        """Validate rules values, before creation"""
        if (hasattr(policy, "rules")):
            for rule in policy["rules"]:
                self.handler.validate_policy_rule(context, policy.id, rule)


def register():
    """Register the NSX-V3 QoS driver."""
    global DRIVER
    if not DRIVER:
        DRIVER = NSXv3QosDriver.create()
    LOG.debug('NSXv3QosDriver QoS driver registered')



# Copyright 2016 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron_lib.api import extensions
from neutron_lib.db import constants as db_const

# The attributes map is here for 2 reasons:
# 1) allow posting id for the different objects we are importing
# 2) make sure security-group named 'default' is also copied

ID_WITH_POST = {'allow_post': True, 'allow_put': False,
                'validate': {'type:uuid': None},
                'is_visible': True,
                'primary_key': True}

RESOURCE_ATTRIBUTE_MAP = {
    'ports': {
        'id': ID_WITH_POST,
    },
    'networks': {
        'id': ID_WITH_POST,
    },
    'security_groups': {
        'id': ID_WITH_POST,
        'name': {'allow_post': True, 'allow_put': True,
                 'is_visible': True, 'default': '',
                 'validate': {'type:string': db_const.NAME_FIELD_SIZE}},
    },
    'security_group_rules': {
        'id': ID_WITH_POST,
    },
    'routers': {
        'id': ID_WITH_POST,
    },
    'policies': {  # QoS policies
        'id': ID_WITH_POST,
    },
}


class Api_replay(extensions.ExtensionDescriptor):
    """Extension for api replay which allows us to specify ids of resources."""

    @classmethod
    def get_name(cls):
        return "Api Replay"

    @classmethod
    def get_alias(cls):
        return 'api-replay'

    @classmethod
    def get_description(cls):
        return "Enables mode to allow api to be replayed"

    @classmethod
    def get_updated(cls):
        return "2016-05-05T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}

    def get_required_extensions(self):
        # make sure this extension is called after those, so our change
        # will not be overridden
        return ["security-group", "router"]

    def get_optional_extensions(self):
        # QoS is optional since it is not always enabled
        return ["qos"]



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import webob.exc

from oslo_utils import uuidutils

from neutron.db import api as db_api
from neutron.db import db_base_plugin_v2
from neutron.db import securitygroups_db
from neutron.extensions import securitygroup as ext_sg
from neutron.tests.unit.extensions import test_securitygroup
from neutron_lib import constants as const
from neutron_lib.plugins import directory

from vmware_nsx.db import extended_security_group_rule as ext_rule_db
from vmware_nsx.extensions import secgroup_rule_local_ip_prefix as ext_loip
from vmware_nsx.plugins.nsx_v.vshield import securitygroup_utils
from vmware_nsx.tests.unit.nsx_v import test_plugin as test_nsxv_plugin
from vmware_nsx.tests.unit.nsx_v3 import test_plugin as test_nsxv3_plugin


PLUGIN_NAME = ('vmware_nsx.tests.unit.extensions.'
               'test_secgroup_rule_local_ip_prefix.ExtendedRuleTestPlugin')

_uuid = uuidutils.generate_uuid


class ExtendedRuleTestPlugin(db_base_plugin_v2.NeutronDbPluginV2,
                             ext_rule_db.ExtendedSecurityGroupRuleMixin,
                             securitygroups_db.SecurityGroupDbMixin):

    supported_extension_aliases = ["security-group",
                                   "secgroup-rule-local-ip-prefix"]

    def create_security_group_rule(self, context, security_group_rule):
        rule = security_group_rule['security_group_rule']
        self._check_local_ip_prefix(context, rule)
        with db_api.context_manager.writer.using(context):
            res = super(ExtendedRuleTestPlugin,
                        self).create_security_group_rule(
                            context, security_group_rule)
            self._process_security_group_rule_properties(context, res, rule)
        return res


class LocalIPPrefixExtTestCase(test_securitygroup.SecurityGroupDBTestCase):
    def setUp(self, plugin=PLUGIN_NAME, ext_mgr=None):
        super(LocalIPPrefixExtTestCase, self).setUp(
            plugin=plugin, ext_mgr=ext_mgr)
        ext_sg.Securitygroup().update_attributes_map(
            ext_loip.RESOURCE_ATTRIBUTE_MAP)

    def _build_ingress_rule_with_local_ip_prefix(self, security_group_id,
                                                 local_ip_prefix,
                                                 remote_ip_prefix,
                                                 direction='ingress'):
        rule = self._build_security_group_rule(
            security_group_id, remote_ip_prefix=remote_ip_prefix,
            direction=direction, proto=const.PROTO_NAME_UDP)
        rule['security_group_rule']['local_ip_prefix'] = local_ip_prefix
        return rule

    def test_raise_rule_not_ingress_when_local_ip_specified(self):
        local_ip_prefix = '239.255.0.0/16'
        remote_ip_prefix = '10.0.0.0/24'
        with self.security_group() as sg:
            rule = self._build_ingress_rule_with_local_ip_prefix(
                sg['security_group']['id'], local_ip_prefix,
                remote_ip_prefix, direction='egress')
            res = self._create_security_group_rule(self.fmt, rule)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_rule_with_local_ip_prefix(self):
        local_ip_prefix = '239.255.0.0/16'
        remote_ip_prefix = '10.0.0.0/24'
        with self.security_group() as sg:
            rule = self._build_ingress_rule_with_local_ip_prefix(
                sg['security_group']['id'], local_ip_prefix, remote_ip_prefix)
            res = self._make_security_group_rule(self.fmt, rule)
            self.assertEqual(local_ip_prefix,
                             res['security_group_rule']['local_ip_prefix'])


class TestNsxVExtendedSGRule(test_nsxv_plugin.NsxVSecurityGroupsTestCase,
                             LocalIPPrefixExtTestCase):
    def test_create_rule_with_local_ip_prefix(self):
        sg_utils = securitygroup_utils.NsxSecurityGroupUtils(None)
        local_ip_prefix = '239.255.0.0/16'
        plugin = directory.get_plugin()
        dest = {'type': 'Ipv4Address', 'value': local_ip_prefix}

        plugin.nsx_sg_utils.get_rule_config = mock.Mock(
            side_effect=sg_utils.get_rule_config)
        super(TestNsxVExtendedSGRule,
              self).test_create_rule_with_local_ip_prefix()
        plugin.nsx_sg_utils.get_rule_config.assert_called_with(
            source=mock.ANY, destination=dest, services=mock.ANY,
            name=mock.ANY, applied_to_ids=mock.ANY, flags=mock.ANY,
            logged=mock.ANY, action=mock.ANY, tag=mock.ANY)


class TestNSXv3ExtendedSGRule(test_nsxv3_plugin.NsxV3PluginTestCaseMixin,
                              LocalIPPrefixExtTestCase):
    def test_create_rule_with_local_ip_prefix(self):
        sg_rules = [
            {'tenant_id': mock.ANY,
             'project_id': mock.ANY,
             'id': mock.ANY,
             'port_range_min': None,
             'local_ip_prefix': '239.255.0.0/16',
             'ethertype': 'IPv4',
             'protocol': u'udp', 'remote_ip_prefix': '10.0.0.0/24',
             'port_range_max': None,
             'security_group_id': mock.ANY,
             'remote_group_id': None, 'direction': u'ingress',
             'description': ''}]

        with mock.patch(
            "vmware_nsxlib.v3.security.NsxLibFirewallSection.create_rules",
            side_effect=test_nsxv3_plugin._mock_create_firewall_rules,
        ) as mock_rule:

            super(TestNSXv3ExtendedSGRule,
                  self).test_create_rule_with_local_ip_prefix()

            mock_rule.assert_called_with(
                mock.ANY,  # content
                mock.ANY,  # firewall_section_id
                mock.ANY,  # ns_group_id
                False,  # logging
                'ALLOW',  # action
                sg_rules,  # sg_rules
                mock.ANY)  # ruleid_2_remote_nsgroup_map



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import mock
from neutron.common import config as neutron_config
from neutron.db import servicetype_db  # noqa
from neutron.quota import resource_registry
from neutron.tests import base
from neutron_lib.callbacks import registry
from neutron_lib.plugins import constants
from oslo_config import cfg
from oslo_log import _options
from oslo_log import log as logging
from oslo_utils import uuidutils
import six
from vmware_nsxlib.v3 import resources as nsx_v3_resources

from vmware_nsx._i18n import _
from vmware_nsx.common import config  # noqa
from vmware_nsx.db import nsxv_db
from vmware_nsx.dvs import dvs_utils
from vmware_nsx.shell.admin.plugins.nsxv.resources import utils as nsxv_utils
from vmware_nsx.shell.admin.plugins.nsxv3.resources import utils as nsxv3_utils
from vmware_nsx.shell import resources
from vmware_nsx.tests import unit as vmware
from vmware_nsx.tests.unit.nsx_v import test_plugin as test_v_plugin
from vmware_nsx.tests.unit.nsx_v3 import test_plugin as test_v3_plugin

LOG = logging.getLogger(__name__)
NSX_INI_PATH = vmware.get_fake_conf('nsx.ini.test')
BASE_CONF_PATH = vmware.get_fake_conf('neutron.conf.test')


@six.add_metaclass(abc.ABCMeta)
class AbstractTestAdminUtils(base.BaseTestCase):

    def setUp(self):
        cfg.CONF.unregister_opts(_options.common_cli_opts)
        cfg.CONF.register_cli_opts(resources.cli_opts)

        super(AbstractTestAdminUtils, self).setUp()

        # remove resource registration conflicts
        resource_registry.unregister_all_resources()

        # Init the neutron config
        neutron_config.init(args=['--config-file', BASE_CONF_PATH,
                                  '--config-file', NSX_INI_PATH])
        self._init_mock_plugin()
        self._init_resource_plugin()
        self.addCleanup(resource_registry.unregister_all_resources)

    def _init_mock_plugin(self):
        mock_query = mock.patch(
            "vmware_nsx.shell.admin.plugins.common.utils.query_yes_no")
        mock_query.start()

    @abc.abstractmethod
    def _get_plugin_name(self):
        pass

    def _init_resource_plugin(self):
        plugin_name = self._get_plugin_name()
        resources.init_resource_plugin(
            plugin_name,
            resources.get_plugin_dir(plugin_name))

    def _test_resource(self, res_name, op, **kwargs):
        errors = self._test_resource_with_errors(res_name, op, **kwargs)
        if len(errors) > 0:
            msg = (_("admin util %(res)s/%(op)s failed with message: "
                     "%(err)s") % {'res': res_name,
                                   'op': op,
                                   'err': errors[0]})
            self.fail(msg=msg)

    def _test_resource_with_errors(self, res_name, op, **kwargs):
        # Must call the internal notify_loop in order to get the errors
        return registry._get_callback_manager()._notify_loop(
            res_name, op, 'nsxadmin', **kwargs)

    def _test_resources(self, res_dict):
        for res in res_dict.keys():
            res_name = res_dict[res].name
            for op in res_dict[res].supported_ops:
                self._test_resource(res_name, op)

    def _test_resources_with_args(self, res_dict, func_args):
        for res in res_dict.keys():
            res_name = res_dict[res].name
            for op in res_dict[res].supported_ops:
                args = {'property': func_args}
                self._test_resource(res_name, op, **args)


class TestNsxvAdminUtils(AbstractTestAdminUtils,
                         test_v_plugin.NsxVPluginV2TestCase):

    def _get_plugin_name(self):
        return 'nsxv'

    def _init_mock_plugin(self, *mocks):
        super(TestNsxvAdminUtils, self)._init_mock_plugin()

        # support the dvs manager:
        mock.patch.object(dvs_utils, 'dvs_create_session').start()
        # override metadata get-object
        dummy_lb = {
            'enabled': True,
            'enableServiceInsertion': True,
            'accelerationEnabled': True,
            'virtualServer': [],
            'applicationProfile': [],
            'pool': [],
            'applicationRule': []
        }
        mock.patch('vmware_nsx.plugins.nsx_v.vshield.nsxv_edge_cfg_obj.'
                   'NsxvEdgeCfgObj.get_object',
                   return_value=dummy_lb).start()

        # Tests shouldn't wait for dummy spawn jobs to finish
        mock.patch('vmware_nsx.shell.admin.plugins.nsxv.resources.utils.'
                   'NsxVPluginWrapper.count_spawn_jobs',
                   return_value=0).start()

        self._plugin = nsxv_utils.NsxVPluginWrapper()

        def get_plugin_mock(alias=constants.CORE):
            if alias in (constants.CORE, constants.L3):
                return self._plugin

        mock.patch("neutron_lib.plugins.directory.get_plugin",
                   side_effect=get_plugin_mock).start()

        # Create a router to make sure we have deployed an edge
        self.router = self.create_router()

    def tearDown(self):
        if self.router and self.router.get('id'):
            edgeapi = nsxv_utils.NeutronDbClient()
            self._plugin.delete_router(edgeapi.context, self.router['id'])
        super(TestNsxvAdminUtils, self).tearDown()

    def test_nsxv_resources(self):
        self._test_resources(resources.nsxv_resources)

    def _test_edge_nsx_update(self, edge_id, params):
        args = {'property': ["edge-id=%s" % edge_id]}
        args['property'].extend(params)
        self._test_resource('edges', 'nsx-update', **args)

    def create_router(self):
        # Create an exclusive router (with an edge)
        tenant_id = uuidutils.generate_uuid()
        data = {'router': {'tenant_id': tenant_id}}
        data['router']['name'] = 'dummy'
        data['router']['admin_state_up'] = True
        data['router']['router_type'] = 'exclusive'

        edgeapi = nsxv_utils.NeutronDbClient()
        return self._plugin.create_router(edgeapi.context, data)

    def get_edge_id(self):
        edgeapi = nsxv_utils.NeutronDbClient()
        bindings = nsxv_db.get_nsxv_router_bindings(edgeapi.context.session)
        for binding in bindings:
            if binding.edge_id:
                return binding.edge_id
        # use a dummy edge
        return "edge-1"

    def test_edge_nsx_updates(self):
        """Test eges/nsx-update utility with different inputs."""
        edge_id = self.get_edge_id()
        self._test_edge_nsx_update(edge_id, ["appliances=true"])
        self._test_edge_nsx_update(edge_id, ["size=compact"])
        self._test_edge_nsx_update(edge_id, ["hostgroup=update"])
        self._test_edge_nsx_update(edge_id, ["hostgroup=all"])
        self._test_edge_nsx_update(edge_id, ["hostgroup=clean"])
        self._test_edge_nsx_update(edge_id, ["highavailability=True"])
        self._test_edge_nsx_update(edge_id, ["resource=cpu", "limit=100"])
        self._test_edge_nsx_update(edge_id, ["syslog-server=1.1.1.1",
                                             "syslog-proto=tcp",
                                             "log-level=debug"])

    def test_bad_args(self):
        args = {'property': ["xxx"]}
        errors = self._test_resource_with_errors(
            'networks', 'nsx-update', **args)
        self.assertEqual(1, len(errors))

    def test_resources_with_common_args(self):
        """Run all nsxv admin utilities with some common arguments

        Using arguments like edge-id which many apis need
        This improves the test coverage
        """
        edge_id = self.get_edge_id()
        args = ["edge-id=%s" % edge_id,
                "router-id=e5b9b249-0034-4729-8ab6-fe4dacaa3a12",
                "policy-id=1",
                "network_id=net-1",
                "net-id=net-1",
                "security-group-id=sg-1",
                "dvs-id=dvs-1",
                "moref=virtualwire-1",
                "teamingpolicy=LACP_ACTIVE"
                ]
        self._test_resources_with_args(
            resources.nsxv_resources, args)

    def test_router_recreate(self):
        # Testing router-recreate separately because it may change the edge-id
        edge_id = self.get_edge_id()
        args = {'property': ["edge-id=%s" % edge_id]}
        self._test_resource('routers', 'nsx-recreate', **args)


class TestNsxv3AdminUtils(AbstractTestAdminUtils,
                          test_v3_plugin.NsxV3PluginTestCaseMixin):

    def _patch_object(self, *args, **kwargs):
        patcher = mock.patch.object(*args, **kwargs)
        patcher.start()
        self._patchers.append(patcher)

    def _init_mock_plugin(self):
        test_v3_plugin._mock_nsx_backend_calls()

        # mock resources
        for cls in (nsx_v3_resources.LogicalPort,
                    nsx_v3_resources.LogicalDhcpServer,
                    nsx_v3_resources.LogicalRouter,
                    nsx_v3_resources.SwitchingProfile):

            self._patch_object(cls, '__init__', return_value=None)
            self._patch_object(cls, 'list', return_value={'results': []})
            self._patch_object(cls, 'get',
                               return_value={'id': uuidutils.generate_uuid()})
            self._patch_object(cls, 'update')

        self._patch_object(nsx_v3_resources.SwitchingProfile,
                           'find_by_display_name',
                           return_value=[{'id': uuidutils.generate_uuid()}])
        super(TestNsxv3AdminUtils, self)._init_mock_plugin()

        self._plugin = nsxv3_utils.NsxV3PluginWrapper()
        mock_nm_get_plugin = mock.patch(
            "neutron_lib.plugins.directory.get_plugin")
        self.mock_nm_get_plugin = mock_nm_get_plugin.start()
        self.mock_nm_get_plugin.return_value = self._plugin

    def _get_plugin_name(self):
        return 'nsxv3'

    def test_nsxv3_resources(self):
        self._test_resources(resources.nsxv3_resources)

    def test_resources_with_common_args(self):
        """Run all nsxv3 admin utilities with some common arguments

        Using arguments like dhcp_profile_uuid which many apis need
        This improves the test coverage
        """
        args = ["dhcp_profile_uuid=e5b9b249-0034-4729-8ab6-fe4dacaa3a12",
                "metadata_proxy_uuid=e5b9b249-0034-4729-8ab6-fe4dacaa3a12",
                "nsx-id=e5b9b249-0034-4729-8ab6-fe4dacaa3a12",
                "availability-zone=default",
                "server-ip=1.1.1.1"
                ]
        # Create some neutron objects for the utilities to run on
        self._create_router()
        with self._create_l3_ext_network() as network:
            with self.subnet(network=network) as subnet:
                with self.port(subnet=subnet):
                    # Run all utilities with backend objects
                    self._test_resources_with_args(
                        resources.nsxv3_resources, args)

    def _create_router(self):
        tenant_id = uuidutils.generate_uuid()
        data = {'router': {'tenant_id': tenant_id}}
        data['router']['name'] = 'dummy'
        data['router']['admin_state_up'] = True

        edgeapi = nsxv_utils.NeutronDbClient()
        return self._plugin.create_router(edgeapi.context, data)


class TestNsxtvdAdminUtils(AbstractTestAdminUtils):

    def _get_plugin_name(self):
        return 'nsxtvd'

    def test_nsxtv_resources(self):
        self._test_resources(resources.nsxtvd_resources)



# Copyright 2017 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_dynamic_routing.db import bgp_db
from neutron_dynamic_routing.extensions import bgp as bgp_ext
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import context as n_context
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
from neutron_lib.services import base as service_base
from oslo_log import log as logging

from vmware_nsx.common import locking
from vmware_nsx.common import nsxv_constants
from vmware_nsx.db import nsxv_db
from vmware_nsx.extensions import edge_service_gateway_bgp_peer as ext_esg
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.nsx import utils as tvd_utils
from vmware_nsx.services.dynamic_routing.nsx_v import driver as nsxv_driver

LOG = logging.getLogger(__name__)
PLUGIN_NAME = bgp_ext.BGP_EXT_ALIAS + '_nsx_svc_plugin'


@tvd_utils.filter_plugins
class NSXBgpPlugin(service_base.ServicePluginBase, bgp_db.BgpDbMixin):
    """BGP service plugin for NSX-V as well as TVD plugins.

    Currently only the nsx-v is supported. other plugins will be refused.
    """

    supported_extension_aliases = [bgp_ext.BGP_EXT_ALIAS,
                                   ext_esg.ESG_BGP_PEER_EXT_ALIAS]

    methods_to_separate = ['get_bgp_speakers',
                           'get_bgp_peers']

    def __init__(self):
        super(NSXBgpPlugin, self).__init__()
        self._core_plugin = directory.get_plugin()

        # initialize the supported drivers (currently only NSX-v)
        self.drivers = {}
        try:
            self.drivers[projectpluginmap.NsxPlugins.NSX_V] = (
                nsxv_driver.NSXvBgpDriver(self))
        except Exception:
            # No driver found
            LOG.warning("NSXBgpPlugin failed to initialize the NSX-V driver")
            self.drivers[projectpluginmap.NsxPlugins.NSX_V] = None

        self._register_callbacks()

    def get_plugin_name(self):
        return PLUGIN_NAME

    def get_plugin_type(self):
        return bgp_ext.BGP_EXT_ALIAS

    def get_plugin_description(self):
        """returns string description of the plugin."""
        return ("BGP dynamic routing service for announcement of next-hops "
                "for project networks, floating IP's, and DVR host routes.")

    def _register_callbacks(self):
        registry.subscribe(self.router_interface_callback,
                           resources.ROUTER_INTERFACE,
                           events.AFTER_CREATE)
        registry.subscribe(self.router_interface_callback,
                           resources.ROUTER_INTERFACE,
                           events.AFTER_DELETE)
        registry.subscribe(self.router_gateway_callback,
                           resources.ROUTER_GATEWAY,
                           events.AFTER_UPDATE)
        registry.subscribe(self.router_gateway_callback,
                           resources.ROUTER_GATEWAY,
                           events.AFTER_DELETE)
        registry.subscribe(self._after_service_edge_create_callback,
                           nsxv_constants.SERVICE_EDGE,
                           events.AFTER_CREATE)
        registry.subscribe(self._before_service_edge_delete_callback,
                           nsxv_constants.SERVICE_EDGE,
                           events.BEFORE_DELETE)

    def _get_driver_by_project(self, context, project):
        # Check if the current project id has a matching driver
        # Currently only NSX-V is supported
        if self._core_plugin.is_tvd_plugin():
            plugin_type = self._core_plugin.get_plugin_type_from_project(
                context, project)
        else:
            plugin_type = self._core_plugin.plugin_type()

        if not self.drivers.get(plugin_type):
            msg = (_("Project %(project)s with plugin %(plugin)s has no "
                     "support for dynamic routing") % {
                'project': project,
                'plugin': plugin_type})
            raise n_exc.InvalidInput(error_message=msg)

        return self.drivers[plugin_type]

    def _get_driver_by_speaker(self, context, bgp_speaker_id):
        try:
            speaker = self.get_bgp_speaker(context, bgp_speaker_id)
        except Exception:
            msg = _("BGP speaker %s could not be found") % bgp_speaker_id
            raise n_exc.BadRequest(resource=bgp_ext.BGP_SPEAKER_RESOURCE_NAME,
                                   msg=msg)
        return self._get_driver_by_project(context, speaker['tenant_id'])

    def create_bgp_speaker(self, context, bgp_speaker):
        driver = self._get_driver_by_project(
            context, bgp_speaker['bgp_speaker']['tenant_id'])
        driver.create_bgp_speaker(context, bgp_speaker)
        return super(NSXBgpPlugin, self).create_bgp_speaker(context,
                                                            bgp_speaker)

    def update_bgp_speaker(self, context, bgp_speaker_id, bgp_speaker):
        driver = self._get_driver_by_speaker(context, bgp_speaker_id)
        with locking.LockManager.get_lock(str(bgp_speaker_id)):
            driver.update_bgp_speaker(context, bgp_speaker_id, bgp_speaker)
            # TBD(roeyc): rolling back changes on edges base class call failed.
            return super(NSXBgpPlugin, self).update_bgp_speaker(
                context, bgp_speaker_id, bgp_speaker)

    def delete_bgp_speaker(self, context, bgp_speaker_id):
        driver = self._get_driver_by_speaker(context, bgp_speaker_id)
        with locking.LockManager.get_lock(str(bgp_speaker_id)):
            driver.delete_bgp_speaker(context, bgp_speaker_id)
            super(NSXBgpPlugin, self).delete_bgp_speaker(context,
                                                         bgp_speaker_id)

    def _add_esg_peer_info(self, context, peer):
        # TODO(asarfaty): only if nsxv driver, or do it in the driver itself
        binding = nsxv_db.get_nsxv_bgp_peer_edge_binding(context.session,
                                                         peer['id'])
        if binding:
            peer['esg_id'] = binding['edge_id']

    def get_bgp_peer(self, context, bgp_peer_id, fields=None):
        peer = super(NSXBgpPlugin, self).get_bgp_peer(context,
                                                      bgp_peer_id, fields)
        if not fields or 'esg_id' in fields:
            self._add_esg_peer_info(context, peer)
        return peer

    def get_bgp_peers_by_bgp_speaker(self, context,
                                     bgp_speaker_id, fields=None):
        ret = super(NSXBgpPlugin, self).get_bgp_peers_by_bgp_speaker(
            context, bgp_speaker_id, fields=fields)
        if fields is None or 'esg_id' in fields:
            for peer in ret:
                self._add_esg_peer_info(context, peer)
        return ret

    def _get_driver_by_peer(self, context, bgp_peer_id):
        try:
            peer = self.get_bgp_peer(context, bgp_peer_id)
        except Exception:
            raise bgp_ext.BgpPeerNotFound(id=bgp_peer_id)
        return self._get_driver_by_project(context, peer['tenant_id'])

    def create_bgp_peer(self, context, bgp_peer):
        driver = self._get_driver_by_project(
            context, bgp_peer['bgp_peer']['tenant_id'])
        driver.create_bgp_peer(context, bgp_peer)
        peer = super(NSXBgpPlugin, self).create_bgp_peer(context, bgp_peer)
        # TODO(asarfaty): only if nsxv driver, or do it in the driver itself
        esg_id = bgp_peer['bgp_peer'].get('esg_id')
        if esg_id:
            nsxv_db.add_nsxv_bgp_peer_edge_binding(context.session, peer['id'],
                                                   esg_id)
            peer['esg_id'] = esg_id
        return peer

    def update_bgp_peer(self, context, bgp_peer_id, bgp_peer):
        driver = self._get_driver_by_peer(context, bgp_peer_id)
        super(NSXBgpPlugin, self).update_bgp_peer(context,
                                                  bgp_peer_id, bgp_peer)
        driver.update_bgp_peer(context, bgp_peer_id, bgp_peer)
        return self.get_bgp_peer(context, bgp_peer_id)

    def delete_bgp_peer(self, context, bgp_peer_id):
        driver = self._get_driver_by_peer(context, bgp_peer_id)
        bgp_peer_info = {'bgp_peer_id': bgp_peer_id}
        bgp_speaker_ids = driver._get_bgp_speakers_by_bgp_peer(
            context, bgp_peer_id)
        for speaker_id in bgp_speaker_ids:
            try:
                self.remove_bgp_peer(context, speaker_id, bgp_peer_info)
            except bgp_ext.BgpSpeakerPeerNotAssociated:
                LOG.debug("Couldn't find bgp speaker %s peer binding while "
                          "deleting bgp peer %s", speaker_id, bgp_peer_id)
        super(NSXBgpPlugin, self).delete_bgp_peer(context, bgp_peer_id)

    def add_bgp_peer(self, context, bgp_speaker_id, bgp_peer_info):
        # speaker & peer must belong to the same driver
        if not bgp_peer_info.get('bgp_peer_id'):
            msg = _("bgp_peer_id must be specified")
            raise n_exc.BadRequest(resource='bgp-peer', msg=msg)
        peer_driver = self._get_driver_by_peer(
            context, bgp_peer_info['bgp_peer_id'])
        speaker_driver = self._get_driver_by_speaker(context, bgp_speaker_id)
        if peer_driver != speaker_driver:
            msg = _("Peer and Speaker must belong to the same plugin")
            raise n_exc.InvalidInput(error_message=msg)
        with locking.LockManager.get_lock(str(bgp_speaker_id)):
            speaker_driver.add_bgp_peer(context,
                                        bgp_speaker_id, bgp_peer_info)
            return super(NSXBgpPlugin, self).add_bgp_peer(context,
                                                          bgp_speaker_id,
                                                          bgp_peer_info)

    def remove_bgp_peer(self, context, bgp_speaker_id, bgp_peer_info):
        driver = self._get_driver_by_speaker(context, bgp_speaker_id)
        with locking.LockManager.get_lock(str(bgp_speaker_id)):
            ret = super(NSXBgpPlugin, self).remove_bgp_peer(
                context, bgp_speaker_id, bgp_peer_info)
            driver.remove_bgp_peer(context, bgp_speaker_id, bgp_peer_info)
            return ret

    def _validate_network_plugin(
        self, context, network_info,
        plugin_type=projectpluginmap.NsxPlugins.NSX_V):
        """Make sure the network belongs to the NSX0-V plugin"""
        if not network_info.get('network_id'):
            msg = _("network_id must be specified")
            raise n_exc.BadRequest(resource=bgp_ext.BGP_SPEAKER_RESOURCE_NAME,
                                   msg=msg)
        net_id = network_info['network_id']
        p = self._core_plugin._get_plugin_from_net_id(context, net_id)
        if p.plugin_type() != plugin_type:
            msg = (_('Network should belong to the %s plugin as the bgp '
                     'speaker') % plugin_type)
            raise n_exc.InvalidInput(error_message=msg)

    def add_gateway_network(self, context, bgp_speaker_id, network_info):
        driver = self._get_driver_by_speaker(context, bgp_speaker_id)
        if self._core_plugin.is_tvd_plugin():
            # The plugin of the network and speaker must be the same
            self._validate_network_plugin(context, network_info)

        with locking.LockManager.get_lock(str(bgp_speaker_id)):
            driver.add_gateway_network(context,
                                       bgp_speaker_id,
                                       network_info)
            return super(NSXBgpPlugin, self).add_gateway_network(
                context, bgp_speaker_id, network_info)

    def remove_gateway_network(self, context, bgp_speaker_id, network_info):
        driver = self._get_driver_by_speaker(context, bgp_speaker_id)
        with locking.LockManager.get_lock(str(bgp_speaker_id)):
            super(NSXBgpPlugin, self).remove_gateway_network(
                context, bgp_speaker_id, network_info)
            driver.remove_gateway_network(context,
                                          bgp_speaker_id,
                                          network_info)

    def get_advertised_routes(self, context, bgp_speaker_id):
        driver = self._get_driver_by_speaker(context, bgp_speaker_id)
        return driver.get_advertised_routes(context, bgp_speaker_id)

    def router_interface_callback(self, resource, event, trigger, **kwargs):
        if not kwargs['network_id']:
            # No GW network, hence no BGP speaker associated
            return

        context = kwargs['context'].elevated()
        router_id = kwargs['router_id']
        subnets = kwargs.get('subnets')
        network_id = kwargs['network_id']
        port = kwargs['port']

        speakers = self._bgp_speakers_for_gateway_network(context,
                                                          network_id)
        for speaker in speakers:
            speaker_id = speaker.id
            with locking.LockManager.get_lock(str(speaker_id)):
                speaker = self.get_bgp_speaker(context, speaker_id)
                driver = self._get_driver_by_project(
                    context, speaker['tenant_id'])
                if network_id not in speaker['networks']:
                    continue
                if event == events.AFTER_CREATE:
                    driver.advertise_subnet(context, speaker_id,
                                            router_id, subnets[0])
                if event == events.AFTER_DELETE:
                    subnet_id = port['fixed_ips'][0]['subnet_id']
                    driver.withdraw_subnet(context, speaker_id,
                                           router_id, subnet_id)

    def router_gateway_callback(self, resource, event, trigger, **kwargs):
        context = kwargs.get('context') or n_context.get_admin_context()
        context = context.elevated()
        router_id = kwargs['router_id']
        network_id = kwargs['network_id']
        speakers = self._bgp_speakers_for_gateway_network(context, network_id)

        for speaker in speakers:
            speaker_id = speaker.id
            driver = self._get_driver_by_project(
                context, speaker['tenant_id'])
            with locking.LockManager.get_lock(str(speaker_id)):
                speaker = self.get_bgp_speaker(context, speaker_id)
                if network_id not in speaker['networks']:
                    continue
                if event == events.AFTER_DELETE:
                    gw_ips = kwargs['gateway_ips']
                    driver.disable_bgp_on_router(context,
                                                 speaker,
                                                 router_id,
                                                 gw_ips[0])
                if event == events.AFTER_UPDATE:
                    updated_port = kwargs['updated_port']
                    router = kwargs['router']
                    driver.process_router_gw_port_update(
                        context, speaker, router, updated_port)

    def _before_service_edge_delete_callback(self, resource, event,
                                             trigger, **kwargs):
        context = kwargs['context'].elevated()
        router = kwargs['router']
        ext_net_id = router.gw_port and router.gw_port['network_id']
        gw_ip = router.gw_port and router.gw_port['fixed_ips'][0]['ip_address']
        edge_id = kwargs.get('edge_id')
        speakers = self._bgp_speakers_for_gateway_network(context, ext_net_id)
        for speaker in speakers:
            driver = self._get_driver_by_project(
                context, speaker['tenant_id'])
            with locking.LockManager.get_lock(speaker.id):
                speaker = self.get_bgp_speaker(context, speaker.id)
                if ext_net_id not in speaker['networks']:
                    continue
                driver.disable_bgp_on_router(context, speaker,
                                             router['id'],
                                             gw_ip, edge_id)

    def _after_service_edge_create_callback(self, resource, event,
                                            trigger, **kwargs):
        context = kwargs['context'].elevated()
        router = kwargs['router']
        ext_net_id = router.gw_port and router.gw_port['network_id']
        speakers = self._bgp_speakers_for_gateway_network(context, ext_net_id)
        for speaker in speakers:
            driver = self._get_driver_by_project(
                context, speaker['tenant_id'])
            with locking.LockManager.get_lock(speaker.id):
                speaker = self.get_bgp_speaker(context, speaker.id)
                if ext_net_id not in speaker['networks']:
                    continue
                driver.enable_bgp_on_router(context, speaker, router['id'])


class NSXvBgpPlugin(NSXBgpPlugin):
    """Defined for backwards compatibility only"""
    pass



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import netaddr
from neutron_lib import constants
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
from oslo_config import cfg
import webob.exc

from vmware_nsx.api_client import exception as api_exc
from vmware_nsx.common import config


class MetaDataTestCase(object):

    def _metadata_setup(self, mode=config.MetadataModes.DIRECT,
                        on_demand=False):
        cfg.CONF.set_override('metadata_mode', mode, self.plugin.cfg_group)
        if hasattr(getattr(cfg.CONF, self.plugin.cfg_group),
                   'metadata_on_demand'):
            cfg.CONF.set_override('metadata_on_demand', on_demand,
                                  self.plugin.cfg_group)

    def _metadata_teardown(self):
        cfg.CONF.set_override('metadata_mode', None, self.plugin.cfg_group)
        if hasattr(getattr(cfg.CONF, self.plugin.cfg_group),
                   'metadata_on_demand'):
            cfg.CONF.set_override('metadata_on_demand', False,
                                  self.plugin.cfg_group)

    def _check_metadata(self, expected_subnets, expected_ports):
        subnets = self._list('subnets')['subnets']
        self.assertEqual(len(subnets), expected_subnets)
        meta_net_id, meta_sub_id = None, None
        meta_cidr = netaddr.IPNetwork('169.254.0.0/16')
        for subnet in subnets:
            cidr = netaddr.IPNetwork(subnet['cidr'])
            if meta_cidr == cidr or meta_cidr in cidr.supernet(16):
                meta_sub_id = subnet['id']
                meta_net_id = subnet['network_id']
                break
        ports = self._list(
            'ports',
            query_params='network_id=%s' % meta_net_id)['ports']
        self.assertEqual(len(ports), expected_ports)
        meta_port_id = ports[0]['id'] if ports else None
        return meta_net_id, meta_sub_id, meta_port_id

    def test_router_add_interface_subnet_with_metadata_access(self):
        self._metadata_setup()
        self.test_router_add_interface_subnet()
        self._metadata_teardown()

    def test_router_add_interface_port_with_metadata_access(self):
        self._metadata_setup()
        self.test_router_add_interface_port()
        self._metadata_teardown()

    def test_router_add_interface_dupsubnet_returns_400_with_metadata(self):
        self._metadata_setup()
        self.test_router_add_interface_dup_subnet1_returns_400()
        self._metadata_teardown()

    def test_router_add_interface_overlapped_cidr_returns_400_with(self):
        self._metadata_setup()
        self.test_router_add_interface_overlapped_cidr_returns_400()
        self._metadata_teardown()

    def test_router_remove_interface_inuse_returns_409_with_metadata(self):
        self._metadata_setup()
        self.test_router_remove_interface_inuse_returns_409()
        self._metadata_teardown()

    def test_router_remove_iface_wrong_sub_returns_400_with_metadata(self):
        self._metadata_setup()
        self.test_router_remove_interface_wrong_subnet_returns_400()
        self._metadata_teardown()

    def test_router_delete_with_metadata_access(self):
        self._metadata_setup()
        self.test_router_delete()
        self._metadata_teardown()

    def test_router_delete_with_port_existed_returns_409_with_metadata(self):
        self._metadata_setup()
        self.test_router_delete_with_port_existed_returns_409()
        self._metadata_teardown()

    def test_delete_port_with_metadata(self):
        self._metadata_setup(config.MetadataModes.INDIRECT)
        with self.subnet() as s:
            with self.port(subnet=s, fixed_ips=[], device_id='1234',
                           device_owner=constants.DEVICE_OWNER_DHCP) as port:
                self._delete('ports', port['port']['id'])
        self._metadata_teardown()

    def test_metadatata_network_created_with_router_interface_add(self):
        self._metadata_setup()
        with mock.patch.object(self._plugin_class, 'schedule_network') as f:
            with self.router() as r:
                with self.subnet() as s:
                    self._router_interface_action('add',
                                                  r['router']['id'],
                                                  s['subnet']['id'],
                                                  None)
                    r_ports = self._list('ports')['ports']
                    self.assertEqual(len(r_ports), 2)
                    ips = []
                    for port in r_ports:
                        ips.extend([netaddr.IPAddress(fixed_ip['ip_address'])
                                    for fixed_ip in port['fixed_ips']])
                    meta_cidr = netaddr.IPNetwork('169.254.0.0/16')
                    self.assertTrue(any([ip in meta_cidr for ip in ips]))
                    # Needed to avoid 409.
                    self._router_interface_action('remove',
                                                  r['router']['id'],
                                                  s['subnet']['id'],
                                                  None)
            # Verify that there has been a schedule_network all for the
            # metadata network
            expected_net_name = 'meta-%s' % r['router']['id']
            found = False
            for call in f.call_args_list:
                # The network data are the last of the positional arguments
                net_dict = call[0][-1]
                if net_dict['name'] == expected_net_name:
                    self.assertFalse(net_dict['port_security_enabled'])
                    self.assertFalse(net_dict['shared'])
                    self.assertFalse(net_dict['tenant_id'])
                    found = True
                    break
            else:
                self.fail("Expected schedule_network call for metadata "
                          "network %s not found" % expected_net_name)
            self.assertTrue(found)
        self._metadata_teardown()

    def test_metadata_network_create_rollback_on_create_subnet_failure(self):
        self._metadata_setup()
        with self.router() as r:
            with self.subnet() as s:
                # Raise a NeutronException (eg: NotFound).
                with mock.patch.object(self._plugin_class,
                                       'create_subnet',
                                       side_effect=n_exc.NotFound):
                    self._router_interface_action(
                        'add', r['router']['id'], s['subnet']['id'], None)
                # Ensure metadata network was removed.
                nets = self._list('networks')['networks']
                self.assertEqual(len(nets), 1)
                # Needed to avoid 409.
                self._router_interface_action('remove',
                                              r['router']['id'],
                                              s['subnet']['id'],
                                              None)
        self._metadata_teardown()

    def test_metadata_network_create_rollback_on_add_rtr_iface_failure(self):
        self._metadata_setup()
        with self.router() as r:
            with self.subnet() as s:
                # Save function being mocked.
                real_func = self._plugin_class.add_router_interface
                plugin_instance = directory.get_plugin()

                # Raise a NeutronException when adding metadata subnet
                # to router.
                def side_effect(*args):
                    if args[-1]['subnet_id'] == s['subnet']['id']:
                        # Do the real thing.
                        return real_func(plugin_instance, *args)
                    # Otherwise raise.
                    raise api_exc.NsxApiException()

                with mock.patch.object(self._plugin_class,
                                       'add_router_interface',
                                       side_effect=side_effect):
                    self._router_interface_action(
                        'add', r['router']['id'], s['subnet']['id'], None)
                # Ensure metadata network was removed.
                nets = self._list('networks')['networks']
                self.assertEqual(len(nets), 1)
                # Needed to avoid 409.
                self._router_interface_action('remove',
                                              r['router']['id'],
                                              s['subnet']['id'],
                                              None)
        self._metadata_teardown()

    def test_metadata_network_removed_with_router_interface_remove(self):
        self._metadata_setup()
        with self.router() as r:
            with self.subnet() as s:
                self._router_interface_action('add', r['router']['id'],
                                              s['subnet']['id'], None)
                meta_net_id, meta_sub_id, meta_port_id = self._check_metadata(
                    expected_subnets=2, expected_ports=1)
                self._router_interface_action('remove', r['router']['id'],
                                              s['subnet']['id'], None)
                self._show('networks', meta_net_id,
                           webob.exc.HTTPNotFound.code)
                self._show('ports', meta_port_id,
                           webob.exc.HTTPNotFound.code)
                self._show('subnets', meta_sub_id,
                           webob.exc.HTTPNotFound.code)
        self._metadata_teardown()

    def test_metadata_network_remove_rollback_on_failure(self):
        self._metadata_setup()
        with self.router() as r:
            with self.subnet() as s:
                self._router_interface_action('add', r['router']['id'],
                                              s['subnet']['id'], None)
                networks = self._list('networks')['networks']
                for network in networks:
                    if network['id'] != s['subnet']['network_id']:
                        meta_net_id = network['id']
                ports = self._list(
                    'ports',
                    query_params='network_id=%s' % meta_net_id)['ports']
                meta_port_id = ports[0]['id']
                # Save function being mocked.
                real_func = self._plugin_class.remove_router_interface
                plugin_instance = directory.get_plugin()

                # Raise a NeutronException when removing metadata subnet
                # from router.
                def side_effect(*args):
                    if args[-1].get('subnet_id') == s['subnet']['id']:
                        # Do the real thing.
                        return real_func(plugin_instance, *args)
                    # Otherwise raise.
                    raise api_exc.NsxApiException()

                with mock.patch.object(self._plugin_class,
                                       'remove_router_interface',
                                       side_effect=side_effect):
                    self._router_interface_action('remove', r['router']['id'],
                                                  s['subnet']['id'], None)
                # Metadata network and subnet should still be there.
                self._show('networks', meta_net_id,
                           webob.exc.HTTPOk.code)
                self._show('ports', meta_port_id,
                           webob.exc.HTTPOk.code)
        self._metadata_teardown()

    def test_metadata_network_with_update_subnet_dhcp_enable(self):
        self._metadata_setup(on_demand=True)
        with self.router() as r:
            # Create a DHCP-disabled subnet.
            with self.subnet(enable_dhcp=False) as s:
                self._router_interface_action('add', r['router']['id'],
                                              s['subnet']['id'], None)
                meta_net_id, meta_sub_id, meta_port_id = self._check_metadata(
                    expected_subnets=2, expected_ports=1)
                # Update subnet to DHCP-enabled.
                data = {'subnet': {'enable_dhcp': True}}
                req = self.new_update_request('subnets', data,
                                              s['subnet']['id'])
                res = self.deserialize(self.fmt, req.get_response(self.api))
                self.assertEqual(True, res['subnet']['enable_dhcp'])
                self._check_metadata(expected_subnets=1, expected_ports=0)
                self._show('networks', meta_net_id,
                           webob.exc.HTTPNotFound.code)
                self._show('ports', meta_port_id,
                           webob.exc.HTTPNotFound.code)
                self._show('subnets', meta_sub_id,
                           webob.exc.HTTPNotFound.code)
        self._metadata_teardown()

    def test_metadata_network_with_update_subnet_dhcp_disable(self):
        self._metadata_setup(on_demand=True)
        with self.router() as r:
            # Create a DHCP-enabled subnet.
            with self.subnet(enable_dhcp=True) as s:
                self._router_interface_action('add', r['router']['id'],
                                              s['subnet']['id'], None)
                self._check_metadata(expected_subnets=1, expected_ports=0)
                # Update subnet to DHCP-disabled.
                data = {'subnet': {'enable_dhcp': False}}
                req = self.new_update_request('subnets', data,
                                              s['subnet']['id'])
                res = self.deserialize(self.fmt, req.get_response(self.api))
                self.assertEqual(False, res['subnet']['enable_dhcp'])
                meta_net_id, meta_sub_id, meta_port_id = self._check_metadata(
                    expected_subnets=2, expected_ports=1)
                self._show('networks', meta_net_id,
                           webob.exc.HTTPOk.code)
                self._show('ports', meta_port_id,
                           webob.exc.HTTPOk.code)
                self._show('subnets', meta_sub_id,
                           webob.exc.HTTPOk.code)
        self._metadata_teardown()

    def test_metadata_dhcp_host_route(self):
        self._metadata_setup(config.MetadataModes.INDIRECT)
        subnets = self._list('subnets')['subnets']
        with self.subnet() as s:
            with self.port(subnet=s, device_id='1234',
                           device_owner=constants.DEVICE_OWNER_DHCP) as port:
                subnets = self._list('subnets')['subnets']
                self.assertEqual(len(subnets), 1)
                subnet_ip_net = netaddr.IPNetwork(s['subnet']['cidr'])
                self.assertIn(netaddr.IPAddress(
                                  subnets[0]['host_routes'][0]['nexthop']),
                              subnet_ip_net)
                self.assertEqual(subnets[0]['host_routes'][0]['destination'],
                                 '169.254.169.254/32')
            self._delete('ports', port['port']['id'])
            subnets = self._list('subnets')['subnets']
            # Test that route is deleted after dhcp port is removed.
            self.assertEqual(len(subnets[0]['host_routes']), 0)
        self._metadata_teardown()
